from __future__ import annotations

from types import MethodType
import builtins
import logging
from typing import (
    Callable,
    Generic,
    Iterator,
    Sequence,
    SupportsIndex,
    overload,
    Any,
    TYPE_CHECKING,
    TypeVar,
    get_type_hints,
    Union,
    Type,
    NoReturn,
    cast,
)
from typing_extensions import get_args, get_origin, ParamSpec, Self
import warnings
import weakref
from contextlib import suppress, contextmanager
from functools import wraps, partial, lru_cache, reduce
import inspect
from inspect import Parameter, Signature, isclass
import threading
import numpy as np

from psygnal import EmitLoopError

from tabulous._range import RectRange, AnyRange, MultiRectRange, TableAnchorBase
from tabulous._selection_op import iter_extract_with_range
from tabulous.exceptions import UnreachableError

__all__ = ["SignalArray"]

logger = logging.getLogger(__name__)
_P = ParamSpec("_P")
_R = TypeVar("_R")

if TYPE_CHECKING:
    from tabulous.widgets._table import _DataFrameTableLayer
    import pandas as pd

    MethodRef = tuple[weakref.ReferenceType[object], str, Union[Callable, None]]
    NormedCallback = Union[MethodRef, Callable]
    StoredSlot = tuple[NormedCallback, Union[int, None]]
    ReducerFunc = Callable[[tuple, tuple], tuple]

    Slice1D = Union[SupportsIndex, slice]
    Slice2D = tuple[Slice1D, Slice1D]

# "safe" builtin functions
# fmt: off
_BUILTINS = {
    k: getattr(builtins, k)
    for k in [
        "int", "str", "float", "bool", "list", "tuple", "set", "dict", "range",
        "slice", "frozenset", "len", "abs", "min", "max", "sum", "any", "all",
        "divmod", "id", "bin", "oct", "hex", "hash", "iter", "isinstance",
        "issubclass", "ord"
    ]
}
# fmt: on


class RangedSlot(Generic[_P, _R], TableAnchorBase):
    """
    Callable object tagged with response range.

    This object will be used in `SignalArray` to store the callback function.
    `range` indicates the range that the callback function will be called.
    """

    def __init__(self, func: Callable[_P, _R], range: RectRange = AnyRange()):
        if not callable(func):
            raise TypeError(f"func must be callable, not {type(func)}")
        if not isinstance(range, RectRange):
            raise TypeError("range must be a RectRange")
        self._func = func
        self._range = range
        wraps(func)(self)

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> Any:
        return self._func(*args, **kwargs)

    def __eq__(self, other: Any) -> bool:
        """Also return True if the wrapped function is the same."""
        if isinstance(other, RangedSlot):
            other = other._func
        return self._func == other

    def __repr__(self) -> str:
        clsname = type(self).__name__
        return f"{clsname}<{self._func!r}>"

    @property
    def range(self) -> RectRange:
        """Slot range."""
        return self._range

    @property
    def func(self) -> Callable[_P, _R]:
        """The wrapped function."""
        return self._func

    def insert_columns(self, col: int, count: int) -> None:
        """Insert columns and update range."""
        return self._range.insert_columns(col, count)

    def insert_rows(self, row: int, count: int) -> None:
        """Insert rows and update range."""
        return self._range.insert_rows(row, count)

    def remove_columns(self, col: int, count: int) -> None:
        """Remove columns and update range."""
        return self._range.remove_columns(col, count)

    def remove_rows(self, row: int, count: int) -> None:
        """Remove rows and update range."""
        return self._range.remove_rows(row, count)


class InCellExpr:
    SELECT = object()

    def __init__(self, objs: list):
        self._objs = objs

    def eval(self, ns: dict[str, Any], ranges: MultiRectRange):
        return self.eval_and_format(ns, ranges)[0]

    def eval_and_format(self, ns: dict[str, Any], ranges: MultiRectRange):
        expr = self.as_literal(ranges)
        logger.debug(f"About to run: {expr!r}")
        ns["__builtins__"] = _BUILTINS
        out = eval(expr, ns, {})
        return out, expr

    def as_literal(self, ranges: MultiRectRange) -> str:
        out: list[str] = []
        _it = iter(ranges)
        for o in self._objs:
            if o is self.SELECT:
                op = next(_it)
                out.append(op.as_iloc_string())
            else:
                out.append(o)
        return "".join(out)


class InCellRangedSlot(RangedSlot[_P, _R]):
    """A slot object with a reference to the table and position."""

    def __init__(
        self,
        expr: InCellExpr,
        pos: tuple[int, int],
        table: _DataFrameTableLayer,
        range: RectRange = AnyRange(),
    ):
        self._expr = expr
        super().__init__(lambda: self.call(), range)
        self._table = weakref.ref(table)
        self._last_destination: tuple[slice, slice] | None = None
        self._current_error: Exception | None = None
        self.set_pos(pos)

    def __repr__(self) -> str:
        expr = self.as_literal()
        return f"{type(self).__name__}<{expr!r}>"

    def as_literal(self, dest: bool = False) -> str:
        """As a literal string that represents this slot."""
        _expr = self._expr.as_literal(self.range)
        if dest:
            if sl := self.last_destination:
                rsl, csl = sl
                _expr = f"df.iloc[{_fmt_slice(rsl)}, {_fmt_slice(csl)}] = {_expr}"
            else:
                _expr = f"out = {_expr}"
        return _expr

    def format_error(self) -> str:
        """Format current exception as a string."""
        if self._current_error is None:
            return ""
        else:
            exc_type = type(self._current_error).__name__
            exc_msg = str(self._current_error)
            return f"{exc_type}: {exc_msg}"

    @property
    def table(self) -> _DataFrameTableLayer:
        """Get the parent table"""
        if table := self._table():
            return table
        raise RuntimeError("Table has been deleted.")

    @property
    def pos(self) -> tuple[int, int]:
        """The visual position of the cell that this slot is attached to."""
        return self._pos

    @property
    def source_pos(self) -> tuple[int, int]:
        """The source position of the cell that this slot is attached to."""
        return self._source_pos

    def set_pos(self, pos: tuple[int, int]):
        """Set the position of the cell that this slot is attached to."""
        self._pos = pos
        prx = self.table.proxy._get_proxy_object()
        self._source_pos = prx.get_source_index(pos[0]), pos[1]
        return self

    @property
    def last_destination(self) -> tuple[slice, slice] | None:
        return self._last_destination

    @last_destination.setter
    def last_destination(self, val):
        if val is None:
            self._last_destination = None
        r, c = val
        if isinstance(r, int):
            r = slice(r, r + 1)
        if isinstance(c, int):
            c = slice(c, c + 1)
        self._last_destination = r, c

    @classmethod
    def from_table(
        cls: type[Self],
        table: _DataFrameTableLayer,
        expr: str,
        pos: tuple[int, int],
    ) -> Self:
        """Construct expression `expr` from `table` at `pos`."""
        qtable = table.native

        # normalize expression to iloc-slicing.
        df_ref = qtable._data_raw
        current_end = 0
        output: list[str] = []
        ranges: list[tuple[slice, slice]] = []
        for (start, end), op in iter_extract_with_range(expr):
            output.append(expr[current_end:start])
            output.append(InCellExpr.SELECT)
            ranges.append(op.as_iloc_slices(df_ref))
            current_end = end
        output.append(expr[current_end:])
        expr_obj = InCellExpr(output)

        # func pos range
        return cls(expr_obj, pos, table, MultiRectRange.from_slices(ranges))

    def evaluate(self) -> EvalResult:
        """Evaluate expression, update cells and return the result."""
        import pandas as pd

        table = self.table
        qtable = table._qwidget
        qtable_view = qtable._qtable_view
        qviewer = qtable.parentViewer()
        self._current_error = None

        df = qtable.getDataFrame()
        if qviewer is not None:
            ns = dict(qviewer._namespace)
        else:
            ns = {"np": np, "pd": pd}
        ns.update(df=df)
        try:
            out, _expr = self._expr.eval_and_format(ns, self.range)
            logger.debug(f"Evaluated at {self.pos!r}")
        except Exception as e:
            logger.debug(f"Evaluation failed at {self.pos!r}: {e!r}")
            self._current_error = e
            return EvalResult(e, self.source_pos)

        _row, _col = self.source_pos

        _is_named_tuple = isinstance(out, tuple) and hasattr(out, "_fields")
        _is_dict = isinstance(out, dict)
        if _is_named_tuple or _is_dict:
            # fmt: off
            with qtable_view._selection_model.blocked(), \
                table.events.data.blocked(), \
                table.proxy.released():
                table.cell.set_labeled_data(_row, _col, out, sep=":")
            # fmt: on

            self.last_destination = (
                slice(_row, _row + len(out)),
                slice(_col, _col + 1),
            )
            return EvalResult(out, (_row, _col))

        if isinstance(out, pd.DataFrame):
            if out.shape[0] > 1 and out.shape[1] == 1:  # 1D array
                _out = out.iloc[:, 0]
                _row, _col = self._infer_slices(_out)
            elif out.size == 1:
                _out = out.iloc[0, 0]
                _row, _col = self._infer_indices()
            else:
                raise NotImplementedError("Cannot assign a DataFrame now.")

        elif isinstance(out, (pd.Series, pd.Index)):
            if out.shape == (1,):  # scalar
                _out = out.values[0]
                _row, _col = self._infer_indices()
            else:  # update a column
                _out = out
                _row, _col = self._infer_slices(_out)

        elif isinstance(out, np.ndarray):
            _out = np.squeeze(out)
            if _out.size == 0:
                raise CellEvaluationError(
                    "Evaluation returned 0-sized array.", self.source_pos
                )
            if _out.ndim == 0:  # scalar
                _out = qtable.convertValue(_col, _out.item())
                _row, _col = self._infer_indices()
            elif _out.ndim == 1:  # 1D array
                _row, _col = self._infer_slices(_out)
            elif _out.ndim == 2:
                _row = slice(_row, _row + _out.shape[0])
                _col = slice(_col, _col + _out.shape[1])
            else:
                raise CellEvaluationError("Cannot assign a >3D array.", self.source_pos)

        else:
            _out = qtable.convertValue(_col, out)

        if isinstance(_row, slice) and isinstance(_col, slice):  # set 1D array
            _out = pd.DataFrame(out).astype(str)
            if _row.start == _row.stop - 1:  # row vector
                _out = _out.T

        elif isinstance(_row, int) and isinstance(_col, int):  # set scalar
            _out = str(_out)

        else:
            raise UnreachableError(type(_row), type(_col))

        _sel_model = qtable_view._selection_model
        with (
            _sel_model.blocked(),
            qtable_view._table_map.lock_pos(self.pos),
            table.undo_manager.merging(lambda _: f"{self.as_literal(dest=True)}"),
            table.proxy.released(keep_widgets=True),
        ):
            qtable.setDataFrameValue(_row, _col, _out)
            qtable.model()._background_color_anim.start(_row, _col)
        self.last_destination = (_row, _col)
        return EvalResult(out, (_row, _col))

    def after_called(self, out: EvalResult) -> None:
        table = self.table
        qtable = table._qwidget
        qtable_view = qtable._qtable_view

        err = out.get_err()

        if err and (sl := self.last_destination):
            import pandas as pd

            rsl, csl = sl
            # determine the error object
            if table.table_type == "SpreadSheet":
                err_repr = "#ERROR"
            else:
                err_repr = pd.NA
            val = np.full(
                (rsl.stop - rsl.start, csl.stop - csl.start),
                err_repr,
                dtype=object,
            )
            # insert error values
            with (
                qtable_view._selection_model.blocked(),
                qtable_view._table_map.lock_pos(self.pos),
                table.events.data.blocked(),
                table.proxy.released(keep_widgets=True),
            ):
                qtable.setDataFrameValue(rsl, csl, pd.DataFrame(val))
                qtable.model()._background_color_anim.start(rsl, csl)
        return None

    def call(self):
        """Function that will be called when cells changed."""
        out = self.evaluate()
        self.after_called(out)
        return out

    def raise_in_msgbox(self, parent=None) -> None:
        """Raise current error in a message box."""
        if self._current_error is None:
            raise ValueError("No error to raise.")
        from tabulous._qt._traceback import QtErrorMessageBox

        return QtErrorMessageBox.from_exc(
            self._current_error, parent=parent
        ).exec_traceback()

    def insert_columns(self, col: int, count: int) -> None:
        """Insert columns and update range."""
        self._range.insert_columns(col, count)
        if dest := self.last_destination:
            rect = RectRange(*dest)
            rect.insert_columns(col, count)
            self.last_destination = rect.as_iloc()
        r, c = self.pos
        if c >= col:
            self.set_pos((r, c + count))

    def insert_rows(self, row: int, count: int) -> None:
        """Insert rows and update range."""
        self._range.insert_rows(row, count)
        if dest := self.last_destination:
            rect = RectRange(*dest)
            rect.insert_rows(row, count)
            self.last_destination = rect.as_iloc()
        r, c = self.pos
        if r >= row:
            self.set_pos((r + count, c))

    def remove_columns(self, col: int, count: int) -> None:
        """Remove columns and update range."""
        self._range.remove_columns(col, count)
        if dest := self.last_destination:
            rect = RectRange(*dest)
            rect.remove_columns(col, count)
            self.last_destination = rect.as_iloc()
        r, c = self.pos
        if c >= col:
            self.set_pos((r, c - count))

    def remove_rows(self, row: int, count: int) -> None:
        """Remove rows and update range."""
        self._range.remove_rows(row, count)
        r, c = self.pos
        if dest := self.last_destination:
            rect = RectRange(*dest)
            rect.remove_rows(row, count)
            self.last_destination = rect.as_iloc()
        if r >= row:
            self.set_pos((r - count, c))

    def _infer_indices(self) -> tuple[int, int]:
        """Infer how to concatenate a scalar to ``df``."""
        #  x | x | x |     1. Self-update is not safe. Raise Error.
        #  x |(1)| x |(2)  2. OK.
        #  x | x | x |     3. OK.
        # ---+---+---+---  4. Cannot determine in which orientation results should
        #    |(3)|   |(4)     be aligned. Raise Error.

        # Filter array selection.
        array_sels = list(self._range.iter_ranges())
        r, c = self.pos

        if len(array_sels) == 0:
            # if no array selection is found, return as a column vector.
            return r, c

        for rloc, cloc in array_sels:
            in_r_range = rloc.start <= r < rloc.stop
            in_c_range = cloc.start <= c < cloc.stop

            if in_r_range and in_c_range:
                raise CellEvaluationError(
                    "Cell evaluation result overlaps with an array selection.",
                    pos=(r, c),
                )
        return r, c

    def _infer_slices(self, out: pd.Series | np.ndarray) -> tuple[slice, slice]:
        """Infer how to concatenate ``out`` to ``df``, based on the selections"""
        #  x | x | x |     1. Self-update is not safe. Raise Error.
        #  x |(1)| x |(2)  2. Return as a column vector.
        #  x | x | x |     3. Return as a row vector.
        # ---+---+---+---  4. Cannot determine in which orientation results should
        #    |(3)|   |(4)     be aligned. Raise Error.

        # Filter array selection.
        array_sels = list(self.range.iter_ranges())
        r, c = self.pos
        len_out = len(out)

        if len(array_sels) == 0:
            # if no array selection is found, return as a column vector.
            return slice(r, r + len_out), slice(c, c + 1)

        determined = None
        for rloc, cloc in array_sels:
            if (
                rloc.stop - rloc.start == 1
                and cloc.stop - cloc.start == 1
                and determined is not None
            ):
                continue
            in_r_range = rloc.start <= r < rloc.stop
            in_c_range = cloc.start <= c < cloc.stop
            r_len = rloc.stop - rloc.start
            c_len = cloc.stop - cloc.start

            if in_r_range:
                if in_c_range:
                    raise CellEvaluationError(
                        "Cell evaluation result overlaps with an array selection.",
                        pos=(r, c),
                    )
                else:
                    if determined is None and len_out <= r_len:
                        determined = (
                            slice(rloc.start, rloc.start + len_out),
                            slice(c, c + 1),
                        )  # column vector

            elif in_c_range:
                if determined is None and len_out <= c_len:
                    determined = (
                        slice(r, r + 1),
                        slice(cloc.start, cloc.start + len_out),
                    )  # row vector
            else:
                # cannot determine output positions, try next selection.
                pass

        if determined is None:
            raise CellEvaluationError(
                "Cell evaluation result is ambiguous. Could not determine the "
                "cells to write output.",
                pos=(r, c),
            )
        return determined


class CellEvaluationError(Exception):
    """Raised when cell evaluation is conducted in a wrong way."""

    def __init__(self, msg: str, pos: tuple[int, int]) -> None:
        super().__init__(msg)
        self._pos = pos


_NULL = object()


class Signal:
    """Copy of psygnal.Signal, without mypyc compilation."""

    __slots__ = (
        "_name",
        "_signature",
        "description",
        "_check_nargs_on_connect",
        "_check_types_on_connect",
    )

    if TYPE_CHECKING:  # pragma: no cover
        _signature: Signature  # callback signature for this signal

    _current_emitter: SignalInstance | None = None

    def __init__(
        self,
        *types: Type[Any] | Signature,
        description: str = "",
        name: str | None = None,
        check_nargs_on_connect: bool = True,
        check_types_on_connect: bool = False,
    ) -> None:

        self._name = name
        self.description = description
        self._check_nargs_on_connect = check_nargs_on_connect
        self._check_types_on_connect = check_types_on_connect

        if types and isinstance(types[0], Signature):
            self._signature = types[0]
            if len(types) > 1:
                warnings.warn(
                    "Only a single argument is accepted when directly providing a"
                    f" `Signature`.  These args were ignored: {types[1:]}"
                )
        else:
            self._signature = _build_signature(*cast("tuple[Type[Any], ...]", types))

    @property
    def signature(self) -> Signature:
        """[Signature][inspect.Signature] supported by this Signal."""
        return self._signature

    def __set_name__(self, owner: Type[Any], name: str) -> None:
        """Set name of signal when declared as a class attribute on `owner`."""
        if self._name is None:
            self._name = name

    def __getattr__(self, name: str) -> Any:
        """Get attribute. Provide useful error if trying to get `connect`."""
        if name == "connect":
            name = self.__class__.__name__
            raise AttributeError(
                f"{name!r} object has no attribute 'connect'. You can connect to the "
                "signal on the *instance* of a class with a Signal() class attribute. "
                "Or create a signal instance directly with SignalInstance."
            )
        return self.__getattribute__(name)

    def __get__(
        self, instance: Any, owner: Type[Any] | None = None
    ) -> Signal | SignalInstance:
        if instance is None:
            return self
        name = cast("str", self._name)
        signal_instance = SignalInstance(
            self.signature,
            instance=instance,
            name=name,
            check_nargs_on_connect=self._check_nargs_on_connect,
            check_types_on_connect=self._check_types_on_connect,
        )
        # instead of caching this signal instance on self, we just assign it
        # to instance.name ... this essentially breaks the descriptor,
        # (i.e. __get__ will never again be called for this instance, and we have no
        # idea how many instances are out there),
        # but it allows us to prevent creating a key for this instance (which may
        # not be hashable or weak-referenceable), and also provides a significant
        # speedup on attribute access (affecting everything).
        setattr(instance, name, signal_instance)
        return signal_instance

    @classmethod
    @contextmanager
    def _emitting(cls, emitter: SignalInstance) -> Iterator[None]:
        """Context that sets the sender on a receiver object while emitting a signal."""
        previous, cls._current_emitter = cls._current_emitter, emitter
        try:
            yield
        finally:
            cls._current_emitter = previous

    @classmethod
    def current_emitter(cls) -> SignalInstance | None:
        """Return currently emitting `SignalInstance`, if any.
        This will typically be used in a callback.
        Examples
        --------
        ```python
        from psygnal import Signal
        def my_callback():
            source = Signal.current_emitter()
        ```
        """
        return cls._current_emitter

    @classmethod
    def sender(cls) -> Any:
        """Return currently emitting object, if any.
        This will typically be used in a callback.
        """
        return getattr(cls._current_emitter, "instance", None)


_empty_signature = Signature()


class SignalInstance:
    """Copy of psygnal.SignalInstance, without mypyc compilation."""

    __slots__ = (
        "_signature",
        "_instance",
        "_name",
        "_slots",
        "_is_blocked",
        "_is_paused",
        "_args_queue",
        "_lock",
        "_check_nargs_on_connect",
        "_check_types_on_connect",
        "__weakref__",
    )

    def __init__(
        self,
        signature: Signature | tuple = _empty_signature,
        *,
        instance: Any = None,
        name: str | None = None,
        check_nargs_on_connect: bool = True,
        check_types_on_connect: bool = False,
    ) -> None:
        self._name = name
        self._instance: Any = instance
        self._args_queue: list[Any] = []  # filled when paused

        if isinstance(signature, (list, tuple)):
            signature = _build_signature(*signature)
        elif not isinstance(signature, Signature):  # pragma: no cover
            raise TypeError(
                "`signature` must be either a sequence of types, or an "
                "instance of `inspect.Signature`"
            )

        self._signature = signature
        self._check_nargs_on_connect = check_nargs_on_connect
        self._check_types_on_connect = check_types_on_connect
        self._slots: list[StoredSlot] = []
        self._is_blocked: bool = False
        self._is_paused: bool = False
        self._lock = threading.RLock()

    @property
    def signature(self) -> Signature:
        """Signature supported by this `SignalInstance`."""
        return self._signature

    @property
    def instance(self) -> Any:
        """Object that emits this `SignalInstance`."""
        return self._instance

    @property
    def name(self) -> str:
        """Name of this `SignalInstance`."""
        return self._name or ""

    def __repr__(self) -> str:
        """Return repr."""
        name = f" {self.name!r}" if self.name else ""
        instance = f" on {self.instance!r}" if self.instance is not None else ""
        return f"<{type(self).__name__}{name}{instance}>"

    def connect(
        self,
        slot: Callable | None = None,
        *,
        check_nargs: bool | None = None,
        check_types: bool | None = None,
        unique: bool | str = False,
        max_args: int | None = None,
    ) -> Callable[[Callable], Callable] | Callable:
        if check_nargs is None:
            check_nargs = self._check_nargs_on_connect
        if check_types is None:
            check_types = self._check_types_on_connect

        def _wrapper(slot: Callable, max_args: int | None = max_args) -> Callable:
            if not callable(slot):
                raise TypeError(f"Cannot connect to non-callable object: {slot}")

            with self._lock:
                if unique and slot in self:
                    if unique == "raise":
                        raise ValueError(
                            "Slot already connect. Use `connect(..., unique=False)` "
                            "to allow duplicate connections"
                        )
                    return slot

                slot_sig = None
                if check_nargs and (max_args is None):
                    slot_sig, max_args = self._check_nargs(slot, self.signature)
                if check_types:
                    slot_sig = slot_sig or signature(slot)
                    if not _parameter_types_match(slot, self.signature, slot_sig):
                        extra = f"- Slot types {slot_sig} do not match types in signal."
                        self._raise_connection_error(slot, extra)

                self._slots.append((_normalize_slot(slot), max_args))
            return slot

        return _wrapper if slot is None else _wrapper(slot)

    def _check_nargs(
        self, slot: Callable, spec: Signature
    ) -> tuple[Signature | None, int | None]:
        """Make sure slot is compatible with signature.
        Also returns the maximum number of arguments that we can pass to the slot
        """
        try:
            slot_sig = _get_signature_possibly_qt(slot)
        except ValueError as e:
            warnings.warn(
                f"{e}. To silence this warning, connect with " "`check_nargs=False`"
            )
            return None, None
        minargs, maxargs = _acceptable_posarg_range(slot_sig)

        n_spec_params = len(spec.parameters)
        # if `slot` requires more arguments than we will provide, raise.
        if minargs > n_spec_params:
            extra = (
                f"- Slot requires at least {minargs} positional "
                f"arguments, but spec only provides {n_spec_params}"
            )
            self._raise_connection_error(slot, extra)
        _sig = None if isinstance(slot_sig, str) else slot_sig
        return _sig, maxargs

    def _raise_connection_error(self, slot: Callable, extra: str = "") -> NoReturn:
        name = getattr(slot, "__name__", str(slot))
        msg = f"Cannot connect slot {name!r} with signature: {signature(slot)}:\n"
        msg += extra
        msg += f"\n\nAccepted signature: {self.signature}"
        raise ValueError(msg)

    def _slot_index(self, slot: NormedCallback) -> int:
        """Get index of `slot` in `self._slots`.  Return -1 if not connected."""
        with self._lock:
            normed = _normalize_slot(slot)
            return next((i for i, s in enumerate(self._slots) if s[0] == normed), -1)

    def disconnect(
        self, slot: NormedCallback | None = None, missing_ok: bool = True
    ) -> None:
        with self._lock:
            if slot is None:
                # NOTE: clearing an empty list is actually a RuntimeError in Qt
                self._slots.clear()
                return

            idx = self._slot_index(slot)
            if idx != -1:
                self._slots.pop(idx)
                if isinstance(slot, PartialMethod):
                    _PARTIAL_CACHE.pop(id(slot), None)
                elif isinstance(slot, tuple) and callable(slot[2]):
                    _prune_partial_cache()
            elif not missing_ok:
                raise ValueError(f"slot is not connected: {slot}")

    def __contains__(self, slot: NormedCallback) -> bool:
        """Return `True` if slot is connected."""
        return self._slot_index(slot) >= 0

    def __len__(self) -> int:
        """Return number of connected slots."""
        return len(self._slots)

    def emit(
        self,
        *args: Any,
        check_nargs: bool = False,
        check_types: bool = False,
        asynchronous: bool = False,
    ) -> EmitThread | None:
        if self._is_blocked:
            return None

        if check_nargs:
            try:
                self.signature.bind(*args)
            except TypeError as e:
                raise TypeError(
                    f"Cannot emit args {args} from signal {self!r} with "
                    f"signature {self.signature}:\n{e}"
                ) from e

        if check_types and not _parameter_types_match(
            lambda: None, self.signature, _build_signature(*(type(a) for a in args))
        ):
            raise TypeError(
                f"Types provided to '{self.name}.emit' "
                f"{tuple(type(a).__name__ for a in args)} do not match signal "
                f"signature: {self.signature}"
            )

        if self._is_paused:
            self._args_queue.append(args)
            return None

        if asynchronous:
            sd = EmitThread(self, args)
            sd.start()
            return sd

        self._run_emit_loop(args)
        return None

    def __call__(
        self,
        *args: Any,
        check_nargs: bool = False,
        check_types: bool = False,
        asynchronous: bool = False,
    ) -> EmitThread | None:
        """Alias for `emit()`."""
        return self.emit(  # type: ignore
            *args,
            check_nargs=check_nargs,
            check_types=check_types,
            asynchronous=asynchronous,
        )

    def _run_emit_loop(self, args: tuple[Any, ...]) -> None:
        rem: list[NormedCallback] = []
        # allow receiver to query sender with Signal.current_emitter()
        with self._lock:
            with Signal._emitting(self):
                for (slot, max_args) in self._slots:
                    if isinstance(slot, tuple):
                        _ref, name, method = slot
                        obj = _ref()
                        if obj is None:
                            rem.append(slot)  # add dead weakref
                            continue
                        if method is not None:
                            cb = method
                        else:
                            _cb = getattr(obj, name, None)
                            if _cb is None:  # pragma: no cover
                                rem.append(slot)  # object has changed?
                                continue
                            cb = _cb
                    else:
                        cb = slot

                    try:
                        cb(*args[:max_args])
                    except Exception as e:
                        raise EmitLoopError(
                            slot=slot, args=args[:max_args], exc=e
                        ) from e

            for slot in rem:
                self.disconnect(slot)

        return None

    def block(self) -> None:
        """Block this signal from emitting."""
        self._is_blocked = True

    def unblock(self) -> None:
        """Unblock this signal, allowing it to emit."""
        self._is_blocked = False

    @contextmanager
    def blocked(self) -> Iterator[None]:
        """Context manager to temporarily block this signal.
        Useful if you need to temporarily block all emission of a given signal,
        (for example, to avoid a recursive signal loop)
        Examples
        --------
        ```python
        class MyEmitter:
            changed = Signal()
            def make_a_change(self):
                self.changed.emit()
        obj = MyEmitter()
        with obj.changed.blocked()
            obj.make_a_change()  # will NOT emit a changed signal.
        ```
        """
        self.block()
        try:
            yield
        finally:
            self.unblock()

    def pause(self) -> None:
        """Pause all emission and collect *args tuples from emit().
        args passed to `emit` will be collected and re-emitted when `resume()` is
        called. For a context manager version, see `paused()`.
        """
        self._is_paused = True

    def resume(self, reducer: ReducerFunc | None = None, initial: Any = _NULL) -> None:
        """Resume (unpause) this signal, emitting everything in the queue.
        Parameters
        ----------
        reducer : Callable[[tuple, tuple], Any], optional
            If provided, all gathered args will be reduced into a single argument by
            passing `reducer` to `functools.reduce`.
            NOTE: args passed to `emit` are collected as tuples, so the two arguments
            passed to `reducer` will always be tuples. `reducer` must handle that and
            return an args tuple.
            For example, three `emit(1)` events would be reduced and re-emitted as
            follows: `self.emit(*functools.reduce(reducer, [(1,), (1,), (1,)]))`
        initial: any, optional
            intial value to pass to `functools.reduce`
        Examples
        --------
        >>> class T:
        ...     sig = Signal(int)
        >>> t = T()
        >>> t.sig.pause()
        >>> t.sig.emit(1)
        >>> t.sig.emit(2)
        >>> t.sig.emit(3)
        >>> t.sig.resume(lambda a, b: (a[0].union(set(b)),), (set(),))
        >>> # results in t.sig.emit({1, 2, 3})
        """
        self._is_paused = False
        # not sure why this attribute wouldn't be set, but when resuming in
        # EventedModel.update, it may be undefined (as seen in tests)
        if not getattr(self, "_args_queue", None):
            return
        if reducer is not None:
            if initial is _NULL:
                args = reduce(reducer, self._args_queue)
            else:
                args = reduce(reducer, self._args_queue, initial)
            self._run_emit_loop(args)
        else:
            for args in self._args_queue:
                self._run_emit_loop(args)
        self._args_queue.clear()

    @contextmanager
    def paused(
        self, reducer: ReducerFunc | None = None, initial: Any = _NULL
    ) -> Iterator[None]:
        """Context manager to temporarly pause this signal.
        Parameters
        ----------
        reducer : Callable[[tuple, tuple], Any], optional
            If provided, all gathered args will be reduced into a single argument by
            passing `reducer` to `functools.reduce`.
            NOTE: args passed to `emit` are collected as tuples, so the two arguments
            passed to `reducer` will always be tuples. `reducer` must handle that and
            return an args tuple.
            For example, three `emit(1)` events would be reduced and re-emitted as
            follows: `self.emit(*functools.reduce(reducer, [(1,), (1,), (1,)]))`
        initial: any, optional
            intial value to pass to `functools.reduce`
        Examples
        --------
        >>> with obj.signal.paused(lambda a, b: (a[0].union(set(b)),), (set(),)):
        ...     t.sig.emit(1)
        ...     t.sig.emit(2)
        ...     t.sig.emit(3)
        >>> # results in obj.signal.emit({1, 2, 3})
        """
        self.pause()
        try:
            yield
        finally:
            self.resume(reducer, initial)

    def __getstate__(self) -> dict:
        """Return dict of current state, for pickle."""
        d = {slot: getattr(self, slot) for slot in self.__slots__}
        d.pop("_lock", None)
        return d


class EmitThread(threading.Thread):
    """A thread to emit a signal asynchronously."""

    def __init__(self, signal_instance: SignalInstance, args: tuple[Any, ...]) -> None:
        super().__init__(name=signal_instance.name)
        self._signal_instance = signal_instance
        self.args = args
        # current = threading.currentThread()
        # self.parent = (current.getName(), current.ident)

    def run(self) -> None:
        """Run thread."""
        self._signal_instance._run_emit_loop(self.args)


# Following codes are mostly copied from psygnal (https://github.com/pyapp-kit/psygnal),
# except for the parametrized part.


class SignalArray(Signal):
    """
    A 2D-parametric signal for a table widget.

    This class is an extension of `psygnal.Signal` that allows partial slot
    connection.

    ```python
    class MyEmitter:
        changed = SignalArray(int)

    emitter = MyEmitter()

    # Connect a slot to the whole table
    emitter.changed.connect(lambda arg: print(arg))
    # Connect a slot to a specific range of the table
    emitter.changed[0:5, 0:4].connect(lambda arg: print("partial:", arg))

    # Emit the signal
    emitter.changed.emit(1)
    # Emit the signal to a specific range
    emitter.changed[8, 8].emit(1)
    ```
    """

    @overload
    def __get__(
        self, instance: None, owner: type[Any] | None = None
    ) -> SignalArray:  # noqa
        ...  # pragma: no cover

    @overload
    def __get__(  # noqa
        self, instance: Any, owner: type[Any] | None = None
    ) -> SignalArrayInstance:
        ...  # pragma: no cover

    def __get__(self, instance: Any, owner: type[Any] | None = None):
        if instance is None:
            return self
        name = self._name
        signal_instance = SignalArrayInstance(
            self.signature,
            instance=instance,
            name=name,
            check_nargs_on_connect=self._check_nargs_on_connect,
            check_types_on_connect=self._check_types_on_connect,
        )
        setattr(instance, name, signal_instance)
        return signal_instance


_empty_signature = Signature()


class SignalArrayInstance(SignalInstance, TableAnchorBase):
    """Parametric version of `SignalInstance`."""

    def __init__(
        self,
        signature: Signature | tuple = _empty_signature,
        *,
        instance: Any = None,
        name: str | None = None,
        check_nargs_on_connect: bool = True,
        check_types_on_connect: bool = False,
    ) -> None:
        super().__init__(
            signature,
            instance=instance,
            name=name,
            check_nargs_on_connect=check_nargs_on_connect,
            check_types_on_connect=check_types_on_connect,
        )

    def __getitem__(self, key: Slice1D | Slice2D) -> _SignalSubArrayRef:
        """Return a sub-array reference."""
        _key = _parse_key(key)
        return _SignalSubArrayRef(self, _key)

    def mloc(self, keys: Sequence[Slice1D | Slice2D]) -> _SignalSubArrayRef:
        ranges = [_parse_key(key) for key in keys]
        return _SignalSubArrayRef(self, MultiRectRange(ranges))

    @overload
    def connect(
        self,
        *,
        check_nargs: bool | None = ...,
        check_types: bool | None = ...,
        unique: bool | str = ...,
        max_args: int | None = None,
        range: RectRange = ...,
    ) -> Callable[[Callable], Callable]:
        ...  # pragma: no cover

    @overload
    def connect(
        self,
        slot: Callable,
        *,
        check_nargs: bool | None = ...,
        check_types: bool | None = ...,
        unique: bool | str = ...,
        max_args: int | None = None,
        range: RectRange = ...,
    ) -> Callable:
        ...  # pragma: no cover

    def connect(
        self,
        slot: Callable | None = None,
        *,
        check_nargs: bool | None = None,
        check_types: bool | None = None,
        unique: bool | str = False,
        max_args: int | None = None,
        range: RectRange = AnyRange(),
    ):
        if check_nargs is None:
            check_nargs = self._check_nargs_on_connect
        if check_types is None:
            check_types = self._check_types_on_connect

        def _wrapper(slot: Callable, max_args: int | None = max_args) -> Callable:
            if not callable(slot):
                raise TypeError(f"Cannot connect to non-callable object: {slot}")

            with self._lock:
                if unique and slot in self:
                    if unique == "raise":
                        raise ValueError(
                            "Slot already connect. Use `connect(..., unique=False)` "
                            "to allow duplicate connections"
                        )
                    return slot

                slot_sig = None
                if check_nargs and (max_args is None):
                    slot_sig, max_args = self._check_nargs(slot, self.signature)
                if check_types:
                    slot_sig = slot_sig or signature(slot)
                    if not _parameter_types_match(slot, self.signature, slot_sig):
                        extra = f"- Slot types {slot_sig} do not match types in signal."
                        self._raise_connection_error(slot, extra)

                self._slots.append((_normalize_slot(RangedSlot(slot, range)), max_args))
            return slot

        return _wrapper if slot is None else _wrapper(slot)

    def connect_cell_slot(
        self,
        slot: InCellRangedSlot,
    ):
        with self._lock:
            _, max_args = self._check_nargs(slot, self.signature)
            self._slots.append((_normalize_slot(slot), max_args))
        return slot

    @overload
    def emit(
        self,
        *args: Any,
        check_nargs: bool = False,
        check_types: bool = False,
        range: RectRange = ...,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def emit(
        self,
        *args: Any,
        check_nargs: bool = False,
        check_types: bool = False,
        range: RectRange = ...,
    ) -> None:
        ...  # pragma: no cover

    def emit(
        self,
        *args: Any,
        check_nargs: bool = False,
        check_types: bool = False,
        range: RectRange = AnyRange(),
    ) -> None:
        if self._is_blocked:
            return None

        if check_nargs:
            try:
                self.signature.bind(*args)
            except TypeError as e:
                raise TypeError(
                    f"Cannot emit args {args} from signal {self!r} with "
                    f"signature {self.signature}:\n{e}"
                ) from e

        if check_types and not _parameter_types_match(
            lambda: None, self.signature, _build_signature(*(type(a) for a in args))
        ):
            raise TypeError(
                f"Types provided to '{self.name}.emit' "
                f"{tuple(type(a).__name__ for a in args)} do not match signal "
                f"signature: {self.signature}"
            )

        if self._is_paused:
            self._args_queue.append(args)
            return None

        self._run_emit_loop(args, range)
        return None

    def insert_rows(self, row: int, count: int) -> None:
        """Insert rows and update slot ranges in-place."""
        for slot, _ in self._slots:
            if isinstance(slot, RangedSlot):
                slot.insert_rows(row, count)
        return None

    def insert_columns(self, col: int, count: int) -> None:
        """Insert columns and update slices in-place."""
        for slot, _ in self._slots:
            if isinstance(slot, RangedSlot):
                slot.insert_columns(col, count)
        return None

    def remove_rows(self, row: int, count: int):
        """Remove rows and update slices in-place."""
        to_be_disconnected: list[RangedSlot] = []
        for slot, _ in self._slots:
            if isinstance(slot, RangedSlot):
                slot.remove_rows(row, count)
                if slot.range.is_empty():
                    logger.debug("Range became empty by removing rows")
                    to_be_disconnected.append(slot)
        for slot in to_be_disconnected:
            self.disconnect(slot, missing_ok=False)
        return None

    def remove_columns(self, col: int, count: int):
        """Remove columns and update slices in-place."""
        to_be_disconnected: list[RangedSlot] = []
        for slot, _ in self._slots:
            if isinstance(slot, RangedSlot):
                slot.remove_columns(col, count)
                if slot.range.is_empty():
                    logger.debug("Range became empty by removing columns")
                    to_be_disconnected.append(slot)
        for slot in to_be_disconnected:
            self.disconnect(slot, missing_ok=False)
        return None

    def _slot_index(self, slot: NormedCallback) -> int:
        """Get index of `slot` in `self._slots`.  Return -1 if not connected."""
        with self._lock:
            if not isinstance(slot, RangedSlot):
                slot = RangedSlot(slot, AnyRange())
            normed = _normalize_slot(slot)
            return next((i for i, s in enumerate(self._slots) if s[0] == normed), -1)

    def _run_emit_loop(
        self,
        args: tuple[Any, ...],
        range: RectRange = AnyRange(),
    ) -> None:
        rem = []

        with self._lock:
            with Signal._emitting(self):
                for (slot, max_args) in self._slots:
                    if isinstance(slot, tuple):
                        _ref, name, method = slot
                        obj = _ref()
                        if obj is None:
                            rem.append(slot)  # add dead weakref
                            continue
                        if method is not None:
                            cb = method
                        else:
                            _cb = getattr(obj, name, None)
                            if _cb is None:  # pragma: no cover
                                rem.append(slot)  # object has changed?
                                continue
                            cb = _cb
                    else:
                        cb = slot

                    if isinstance(cb, RangedSlot) and not range.overlaps_with(cb.range):
                        continue
                    try:
                        cb(*args[:max_args])
                    except Exception as e:
                        raise EmitLoopError(repr(slot), args[:max_args], e) from e

            for slot in rem:
                self.disconnect(slot)

        return None

    def iter_slots(self) -> Iterator[Callable]:
        """Iterate over all connected slots."""
        for slot, _ in self._slots:
            if isinstance(slot, tuple):
                _ref, name, method = slot
                obj = _ref()
                if obj is None:
                    continue
                if method is not None:
                    cb = method
                else:
                    _cb = getattr(obj, name, None)
                    if _cb is None:
                        continue
                    cb = _cb
            else:
                cb = slot
            yield cb


class _SignalSubArrayRef:
    """A reference to a subarray of a signal."""

    def __init__(self, sig: SignalArrayInstance, key):
        self._sig: weakref.ReferenceType[SignalArrayInstance] = weakref.ref(sig)
        self._key = key

    def _get_parent(self) -> SignalArrayInstance:
        sig = self._sig()
        if sig is None:
            raise RuntimeError("Parent SignalArrayInstance has been garbage collected")
        return sig

    def connect(
        self,
        slot: Callable,
        *,
        check_nargs: bool | None = None,
        check_types: bool | None = None,
        unique: bool | str = False,
        max_args: int | None = None,
    ):
        return self._get_parent().connect(
            slot,
            check_nargs=check_nargs,
            check_types=check_types,
            unique=unique,
            max_args=max_args,
            range=self._key,
        )

    def emit(
        self,
        *args: Any,
        check_nargs: bool = False,
        check_types: bool = False,
    ):
        return self._get_parent().emit(
            *args, check_nargs=check_nargs, check_types=check_types, range=self._key
        )


def _parse_a_key(k):
    if isinstance(k, slice):
        return k
    elif isinstance(k, (list, np.ndarray)):
        # fancy slicing, which occurs when the table is filtered/sorted.
        return slice(np.min(k), np.max(k) + 1)
    else:
        k = k.__index__()
        return slice(k, k + 1)


def _parse_key(key):
    if isinstance(key, tuple):
        if len(key) == 2:
            r, c = key
            key = RectRange(_parse_a_key(r), _parse_a_key(c))
        elif len(key) == 1:
            key = RectRange(_parse_a_key(key[0]))
        else:
            raise IndexError("too many indices")
    else:
        key = RectRange(_parse_a_key(key), slice(None))
    return key


def _fmt_slice(sl: slice) -> str:
    s0 = sl.start if sl.start is not None else ""
    s1 = sl.stop if sl.stop is not None else ""
    return f"{s0}:{s1}"


_T = TypeVar("_T")


class EvalResult(Generic[_T]):
    """A Rust-like Result type for evaluation."""

    def __init__(self, obj: _T | Exception, range: tuple[int | slice, int | slice]):
        # TODO: range should be (int, int).
        self._obj = obj
        _r, _c = range
        if isinstance(_r, int):
            _r = slice(_r, _r + 1)
        if isinstance(_c, int):
            _c = slice(_c, _c + 1)
        self._range = (_r, _c)

    def __repr__(self) -> str:
        cname = type(self).__name__
        if isinstance(self._obj, Exception):
            desc = "Err"
        else:
            desc = "Ok"
        return f"{cname}<{desc}({self._obj!r})>"

    def _short_repr(self) -> str:
        cname = type(self).__name__
        if isinstance(self._obj, Exception):
            desc = "Err"
        else:
            desc = "Ok"
        _obj = repr(self._obj)
        if "\n" in _obj:
            _obj = _obj.split("\n")[0] + "..."
        if len(_obj.rstrip("...")) > 20:
            _obj = _obj[:20] + "..."
        return f"{cname}<{desc}({_obj})>"

    @property
    def range(self) -> tuple[slice, slice]:
        """Output range."""
        return self._range

    def unwrap(self) -> _T:
        obj = self._obj
        if isinstance(obj, Exception):
            raise obj
        return obj

    def get_err(self) -> Exception | None:
        if isinstance(self._obj, Exception):
            return self._obj
        return None

    def is_err(self) -> bool:
        """True is an exception is wrapped."""
        return isinstance(self._obj, Exception)


class PartialMethodMeta(type):
    def __instancecheck__(cls, inst: object) -> bool:
        return isinstance(inst, partial) and isinstance(inst.func, MethodType)


class PartialMethod(metaclass=PartialMethodMeta):
    """Bound method wrapped in partial: `partial(MyClass().some_method, y=1)`."""

    func: MethodType
    args: tuple
    keywords: dict[str, Any]


def signature(obj: Any) -> inspect.Signature:
    try:
        return inspect.signature(obj)
    except ValueError as e:
        with suppress(Exception):
            if not inspect.ismethod(obj):
                return _stub_sig(obj)
        raise e from e


@lru_cache(maxsize=None)
def _stub_sig(obj: Any) -> Signature:
    import builtins

    if obj is builtins.print:
        params = [
            Parameter(name="value", kind=Parameter.VAR_POSITIONAL),
            Parameter(name="sep", kind=Parameter.KEYWORD_ONLY, default=" "),
            Parameter(name="end", kind=Parameter.KEYWORD_ONLY, default="\n"),
            Parameter(name="file", kind=Parameter.KEYWORD_ONLY, default=None),
            Parameter(name="flush", kind=Parameter.KEYWORD_ONLY, default=False),
        ]
        return Signature(params)
    raise ValueError("unknown object")


# def f(a, /, b, c=None, *d, f=None, **g): print(locals())
#
# a: kind=POSITIONAL_ONLY,       default=Parameter.empty    # 1 required posarg
# b: kind=POSITIONAL_OR_KEYWORD, default=Parameter.empty    # 1 requires posarg
# c: kind=POSITIONAL_OR_KEYWORD, default=None               # 1 optional posarg
# d: kind=VAR_POSITIONAL,        default=Parameter.empty    # N optional posargs
# e: kind=KEYWORD_ONLY,          default=Parameter.empty    # 1 REQUIRED kwarg
# f: kind=KEYWORD_ONLY,          default=None               # 1 optional kwarg
# g: kind=VAR_KEYWORD,           default=Parameter.empty    # N optional kwargs


def _parameter_types_match(
    function: Callable, spec: Signature, func_sig: Signature | None = None
) -> bool:
    """Return True if types in `function` signature match those in `spec`.

    Parameters
    ----------
    function : Callable
        A function to validate
    spec : Signature
        The Signature against which the `function` should be validated.
    func_sig : Signature, optional
        Signature for `function`, if `None`, signature will be inspected.
        by default None

    Returns
    -------
    bool
        True if the parameter types match.
    """
    fsig = func_sig or signature(function)

    func_hints = None
    for f_param, spec_param in zip(fsig.parameters.values(), spec.parameters.values()):
        f_anno = f_param.annotation
        if f_anno is fsig.empty:
            # if function parameter is not type annotated, allow it.
            continue

        if isinstance(f_anno, str):
            if func_hints is None:
                func_hints = get_type_hints(function)
            f_anno = func_hints.get(f_param.name)

        if not _is_subclass(f_anno, spec_param.annotation):
            return False
    return True


def _is_subclass(left: type[Any], right: type) -> bool:
    """Variant of issubclass with support for unions."""
    if not isclass(left) and get_origin(left) is Union:
        return any(issubclass(i, right) for i in get_args(left))
    return issubclass(left, right)


def _get_method_name(slot: MethodType) -> tuple[weakref.ref, str]:
    obj = slot.__self__
    # some decorators will alter method.__name__, so that obj.method
    # will not be equal to getattr(obj, obj.method.__name__).
    # We check for that case here and find the proper name in the function's closures
    if getattr(obj, slot.__name__, None) != slot:
        for c in slot.__closure__ or ():
            cname = getattr(c.cell_contents, "__name__", None)
            if cname and getattr(obj, cname, None) == slot:
                return weakref.ref(obj), cname
        # slower, but catches cases like assigned functions
        # that won't have function in closure
        for name in reversed(dir(obj)):  # most dunder methods come first
            if getattr(obj, name) == slot:
                return weakref.ref(obj), name
        # we don't know what to do here.
        raise RuntimeError(  # pragma: no cover
            f"Could not find method on {obj} corresponding to decorated function {slot}"
        )
    return weakref.ref(obj), slot.__name__


# #############################################################################
# #############################################################################


def _build_signature(*types: Type[Any]) -> Signature:
    params = [
        Parameter(name=f"p{i}", kind=Parameter.POSITIONAL_ONLY, annotation=t)
        for i, t in enumerate(types)
    ]
    return Signature(params)


def _normalize_slot(slot: Callable | NormedCallback) -> NormedCallback:
    if isinstance(slot, MethodType):
        return _get_method_name(slot) + (None,)
    if isinstance(slot, PartialMethod):
        return _partial_weakref(slot)
    if isinstance(slot, tuple) and not isinstance(slot[0], weakref.ref):
        return (weakref.ref(slot[0]), slot[1], slot[2])
    return slot


# def f(a, /, b, c=None, *d, f=None, **g): print(locals())
#
# a: kind=POSITIONAL_ONLY,       default=Parameter.empty    # 1 required posarg
# b: kind=POSITIONAL_OR_KEYWORD, default=Parameter.empty    # 1 requires posarg
# c: kind=POSITIONAL_OR_KEYWORD, default=None               # 1 optional posarg
# d: kind=VAR_POSITIONAL,        default=Parameter.empty    # N optional posargs
# e: kind=KEYWORD_ONLY,          default=Parameter.empty    # 1 REQUIRED kwarg
# f: kind=KEYWORD_ONLY,          default=None               # 1 optional kwarg
# g: kind=VAR_KEYWORD,           default=Parameter.empty    # N optional kwargs


def _get_signature_possibly_qt(slot: Callable) -> Signature | str:
    # checking qt has to come first, since the signature of the emit method
    # of a Qt SignalInstance is just <Signature (*args: typing.Any) -> None>
    # https://bugreports.qt.io/browse/PYSIDE-1713
    sig = _guess_qtsignal_signature(slot)
    return signature(slot) if sig is None else sig


def _acceptable_posarg_range(
    sig: Signature | str, forbid_required_kwarg: bool = True
) -> tuple[int, int | None]:
    """Return tuple of (min, max) accepted positional arguments.
    Parameters
    ----------
    sig : Signature
        Signature object to evaluate
    forbid_required_kwarg : Optional[bool]
        Whether to allow required KEYWORD_ONLY parameters. by default True.
    Returns
    -------
    arg_range : Tuple[int, int]
        minimum, maximum number of acceptable positional arguments
    Raises
    ------
    ValueError
        If the signature has a required keyword_only parameter and
        `forbid_required_kwarg` is `True`.
    """
    if isinstance(sig, str):
        assert "(" in sig, f"Unrecognized string signature format: {sig}"
        inner = sig.split("(", 1)[1].split(")", 1)[0]
        minargs = maxargs = inner.count(",") + 1 if inner else 0
        return minargs, maxargs

    required = 0
    optional = 0
    posargs_unlimited = False
    _pos_required = {Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD}
    for param in sig.parameters.values():
        if param.kind in _pos_required:
            if param.default is Parameter.empty:
                required += 1
            else:
                optional += 1
        elif param.kind is Parameter.VAR_POSITIONAL:
            posargs_unlimited = True
        elif (
            param.kind is Parameter.KEYWORD_ONLY
            and param.default is Parameter.empty
            and forbid_required_kwarg
        ):
            raise ValueError("Required KEYWORD_ONLY parameters not allowed")
    return (required, None if posargs_unlimited else required + optional)


def _parameter_types_match(
    function: Callable, spec: Signature, func_sig: Signature | None = None
) -> bool:
    """Return True if types in `function` signature match those in `spec`.

    Parameters
    ----------
    function : Callable
        A function to validate
    spec : Signature
        The Signature against which the `function` should be validated.
    func_sig : Signature, optional
        Signature for `function`, if `None`, signature will be inspected.
        by default None

    Returns
    -------
    bool
        True if the parameter types match.
    """
    fsig = func_sig or signature(function)

    func_hints = None
    for f_param, spec_param in zip(fsig.parameters.values(), spec.parameters.values()):
        f_anno = f_param.annotation
        if f_anno is fsig.empty:
            # if function parameter is not type annotated, allow it.
            continue

        if isinstance(f_anno, str):
            if func_hints is None:
                func_hints = get_type_hints(function)
            f_anno = func_hints.get(f_param.name)

        if not _is_subclass(f_anno, spec_param.annotation):
            return False
    return True


_PARTIAL_CACHE: dict[int, tuple[weakref.ref, str, Callable]] = {}


def _partial_weakref(slot_partial: PartialMethod) -> tuple[weakref.ref, str, Callable]:
    """For partial methods, make the weakref point to the wrapped object."""
    _id = id(slot_partial)

    # if the exact same partial is used twice, we don't want to recreate a new
    # wrap() function, because we want _partial_weakref(cb) == _partial_weakref(cb)
    # to be True.  So we cache the result of the first call using the id of the partial
    if _id not in _PARTIAL_CACHE:
        ref, name = _get_method_name(slot_partial.func)
        args_ = slot_partial.args
        kwargs_ = slot_partial.keywords

        def wrap(*args: Any, **kwargs: Any) -> Any:
            getattr(ref(), name)(*args_, *args, **kwargs_, **kwargs)

        _PARTIAL_CACHE[_id] = (ref, name, wrap)
    return _PARTIAL_CACHE[_id]


def _prune_partial_cache() -> None:
    """Remove any partial methods whose object has been garbage collected."""
    for key, (ref, *_) in list(_PARTIAL_CACHE.items()):
        if ref() is None:
            del _PARTIAL_CACHE[key]


def _get_method_name(slot: MethodType) -> tuple[weakref.ref, str]:
    obj = slot.__self__
    # some decorators will alter method.__name__, so that obj.method
    # will not be equal to getattr(obj, obj.method.__name__).
    # We check for that case here and find the proper name in the function's closures
    if getattr(obj, slot.__name__, None) != slot:
        for c in slot.__closure__ or ():
            cname = getattr(c.cell_contents, "__name__", None)
            if cname and getattr(obj, cname, None) == slot:
                return weakref.ref(obj), cname
        # slower, but catches cases like assigned functions
        # that won't have function in closure
        for name in reversed(dir(obj)):  # most dunder methods come first
            if getattr(obj, name) == slot:
                return weakref.ref(obj), name
        # we don't know what to do here.
        raise RuntimeError(  # pragma: no cover
            f"Could not find method on {obj} corresponding to decorated function {slot}"
        )
    return weakref.ref(obj), slot.__name__


def _guess_qtsignal_signature(obj: Any) -> str | None:
    """Return string signature if `obj` is a SignalInstance or Qt emit method.
    This is a bit of a hack, but we found no better way:
    https://stackoverflow.com/q/69976089/1631624
    https://bugreports.qt.io/browse/PYSIDE-1713
    """
    # on my machine, this takes ~700ns on PyQt5 and 8.7µs on PySide2
    type_ = type(obj)
    if "pyqtBoundSignal" in type_.__name__:
        return cast("str", obj.signal)
    qualname = getattr(obj, "__qualname__", "")
    if qualname == "pyqtBoundSignal.emit":
        return cast("str", obj.__self__.signal)
    if qualname == "SignalInstance.emit" and type_.__name__.startswith("builtin"):
        # we likely have the emit method of a SignalInstance
        # call it with ridiculous params to get the err
        return _ridiculously_call_emit(obj.__self__.emit)
    if "SignalInstance" in type_.__name__ and "QtCore" in getattr(
        type_, "__module__", ""
    ):
        return _ridiculously_call_emit(obj.emit)
    return None


_CRAZY_ARGS = (1,) * 255


def _ridiculously_call_emit(emitter: Any) -> str | None:
    """Call SignalInstance emit() to get the signature from err message."""
    try:
        emitter(*_CRAZY_ARGS)
    except TypeError as e:
        if "only accepts" in str(e):
            return str(e).split("only accepts")[0].strip()
    return None  # pragma: no cover



from __future__ import annotations

from qtpy import QtWidgets as QtW, QtGui
from superqt.utils import QSignalThrottler

from tabulous._qt._mainwindow import QMainWindow
from tabulous._qt._qt_const import MonospaceFontFamily
from tabulous._utils import get_config, update_config
from ._shared import QTitleLabel


class QTableConfigPanel(QtW.QWidget):
    def __init__(self, parent: QtW.QWidget | None = None) -> None:
        super().__init__(parent)
        self._throttler = QSignalThrottler(parent=self)
        self._throttler.setTimeout(1000)
        self._throttler.triggered.connect(self._on_throttled)
        self._setup_ui()

    def get_updated_config(self):
        cfg = get_config()
        cfg.table.max_row_count = self._max_row_count.value()
        cfg.table.max_column_count = self._max_column_count.value()
        cfg.table.font = self._font.currentFont().family()
        cfg.table.font_size = self._font_size.value()
        cfg.table.row_size = self._row_size.value()
        cfg.table.column_size = self._column_size.value()
        cfg.cell.ref_prefix = self._ref_prefix.text()
        cfg.cell.eval_prefix = self._eval_prefix.text()
        return cfg

    def _on_throttled(self):
        cfg = self.get_updated_config()
        update_config(cfg)
        QMainWindow.reload_config()

    def _setup_ui(self):
        from tabulous._utils import get_config

        cfg = get_config()

        self._max_row_count = QtW.QSpinBox()
        self._max_row_count.setRange(10_000, 1_000_000)
        self._max_row_count.setValue(cfg.table.max_row_count)

        self._max_column_count = QtW.QSpinBox()
        self._max_column_count.setRange(1_000, 100_000)
        self._max_row_count.setValue(cfg.table.max_column_count)

        self._font = QtW.QFontComboBox()
        self._font.setCurrentFont(QtGui.QFont(cfg.table.font))

        self._font_size = QtW.QSpinBox()
        self._font_size.setRange(6, 28)
        self._font_size.setValue(cfg.table.font_size)

        self._row_size = QtW.QSpinBox()
        self._row_size.setRange(10, 100)
        self._row_size.setValue(cfg.table.row_size)

        self._column_size = QtW.QSpinBox()
        self._column_size.setRange(20, 200)
        self._column_size.setValue(cfg.table.column_size)

        self._ref_prefix = QtW.QLineEdit()
        self._ref_prefix.setFont(QtGui.QFont(MonospaceFontFamily))
        self._ref_prefix.setText(cfg.cell.ref_prefix)

        self._eval_prefix = QtW.QLineEdit()
        self._eval_prefix.setFont(QtGui.QFont(MonospaceFontFamily))
        self._eval_prefix.setText(cfg.cell.eval_prefix)

        _layout = QtW.QVBoxLayout()
        _layout.addWidget(QTitleLabel("Table", 14))
        _widget_table = QtW.QWidget()
        _layout_table = QtW.QFormLayout()
        _layout_table.setContentsMargins(0, 0, 0, 0)
        _widget_table.setLayout(_layout_table)
        _layout.addWidget(_widget_table)
        _layout_table.addRow("Max row count", self._max_row_count)
        _layout_table.addRow("Max column count", self._max_column_count)
        _layout_table.addRow("Font family", self._font)
        _layout_table.addRow("Font size", self._font_size)
        _layout_table.addRow("Default row span", self._row_size)
        _layout_table.addRow("Default column span", self._column_size)
        _layout.addWidget(QTitleLabel("Cell", 14))
        _widget_cell = QtW.QWidget()
        _layout_cell = QtW.QFormLayout()
        _layout_cell.setContentsMargins(0, 0, 0, 0)
        _widget_cell.setLayout(_layout_cell)
        _layout.addWidget(_widget_cell)
        _layout_cell.addRow("Reference prefix", self._ref_prefix)
        _layout_cell.addRow("Evaluation prefix", self._eval_prefix)

        self._max_row_count.valueChanged.connect(self._throttler.throttle)
        self._max_column_count.valueChanged.connect(self._throttler.throttle)
        self._font.currentFontChanged.connect(self._throttler.throttle)
        self._font_size.valueChanged.connect(self._throttler.throttle)
        self._row_size.valueChanged.connect(self._throttler.throttle)
        self._column_size.valueChanged.connect(self._throttler.throttle)
        self._ref_prefix.editingFinished.connect(self._throttler.throttle)
        self._eval_prefix.editingFinished.connect(self._throttler.throttle)

        self.setLayout(_layout)



from __future__ import annotations

from typing import NamedTuple, Protocol
import numpy as np
import pandas as pd

from magicgui import magicgui
from magicgui.widgets import (
    Container,
    Dialog,
    PushButton,
    ComboBox,
    Label,
    TextEdit,
)

from tabulous._magicgui import find_current_table, SelectionWidget, ToggleSwitch
from ._models import MODELS, ADVANCED
from tabulous._qt._qt_const import MonospaceFontFamily


class SkLearnInput(NamedTuple):
    """Input for sklearn algorithms"""

    X: np.ndarray
    Y: np.ndarray | None
    labels: list[str]


class XYContainer(Container):
    """Data selection widgets for sklearn algorithms."""

    def __init__(self, **kwargs):
        self._X_widget = SelectionWidget(format="iloc", name="X")
        self._Y_widget = SelectionWidget(format="iloc", name="Y")
        super().__init__(widgets=[self._X_widget, self._Y_widget], **kwargs)
        self.margins = (0, 0, 0, 0)

    def get_values(self, df: pd.DataFrame) -> SkLearnInput:
        df_sub = self._X_widget.value.operate(df)
        X = df_sub.values
        labels = df_sub.columns.tolist()
        if self._Y_widget.value is not None:
            Y = self._Y_widget.value.operate(df)

            if Y.shape[1] != 1:
                raise ValueError("Label must be a single column")
            if Y.shape[0] != X.shape[0]:
                raise ValueError("Label and data must have the same number of rows")
            Y = Y.values.ravel()
            if Y.dtype.kind not in "ui":
                lut, Y = np.unique(Y, return_inverse=True)
        else:
            Y = None

        return SkLearnInput(X, Y, labels)


class SkLearnModelProtocol(Protocol):
    """A protocol for sklearn models"""

    # fmt: off
    def fit(self, X: np.ndarray, Y: np.ndarray | None = None): ...
    def predict(self, X: np.ndarray) -> np.ndarray: ...
    def fit_predict(self, X: np.ndarray, Y: np.ndarray | None = None) -> np.ndarray: ...
    def transform(self, X: np.ndarray) -> np.ndarray: ...
    def score(self, X: np.ndarray, Y: np.ndarray | None = None) -> float: ...
    def get_params(self, deep: bool = True) -> dict: ...
    # fmt: on


class SkLearnModelEdit(Container):
    """Widget for editing sklearn models and constructing them."""

    def __init__(self, **kwargs):
        self._text = Label(name="Model", value="")
        self._text.min_width = 200
        self._btn = PushButton(text="Select model")
        self._check = ToggleSwitch(
            value=False, label="Advanced", tooltip="Check to show more models"
        )
        self._model: SkLearnModelProtocol | None = None

        super().__init__(
            widgets=[self._text, self._btn, self._check],
            labels=False,
            layout="horizontal",
            **kwargs,
        )
        # disconnect existing signales
        self._text.changed.disconnect()
        self._btn.changed.disconnect()
        self._check.changed.disconnect()
        self._btn.changed.connect(self._on_clicked)
        self._check.changed.connect(self._on_check_changed)

    @property
    def model(self) -> SkLearnModelProtocol | None:
        return self._model

    def _on_clicked(self):
        _model_choice = ComboBox(choices=MODELS.keys(), nullable=False)
        if model_name := self._text.value:
            model_name: str
            if model_name.endswith(" (fitted)"):
                model_name = model_name[:-9]
            _model_choice.value = model_name
        container = Container(widgets=[], labels=False)
        container.margins = (0, 0, 0, 0)

        dlg = Dialog(widgets=[_model_choice, container])

        @_model_choice.changed.connect
        def _on_model_changed():
            if model_name := _model_choice.value:
                model_factory = MODELS[model_name]
                gui = magicgui(model_factory, call_button=False)
                if len(container) > 0:
                    container.clear()
                container.append(gui)

        _model_choice.changed.emit(_model_choice.value)
        if dlg.exec():
            mgui = dlg[-1][0]
            self._model = mgui()
            self._text.value = _model_choice.value
            self.changed.emit(self._model)
        return None

    def _on_check_changed(self, checked: bool):
        if checked:
            MODELS.show_keys()  # advanced mode
        else:
            MODELS.hide_keys(ADVANCED)  # basic mode
        return None


class SkLearnContainer(Container):
    """
    The main container widget for scikit-learn analysis.

    This widget contains a data selection widget, a model selection widget, and
    buttons that perform the analysis.
    """

    _current_widget = None

    def __init__(self, **kwargs):
        self._model_widget = SkLearnModelEdit(label="model")
        self._data_widget = XYContainer(label="data")

        self._fit_button = PushButton(label="Fit", tooltip="Fit model by data")
        self._predict_button = PushButton(
            label="Predict", tooltip="Predict labels by data"
        )
        self._fit_predict_button = PushButton(
            label="Fit/Predict", tooltip="Fit model and predict labels by data"
        )
        self._transform_button = PushButton(label="Transform", tooltip="Transform data")
        self._describe_button = PushButton(
            label="Describe", tooltip="Describe the model"
        )
        self._score_button = PushButton(label="Score", tooltip="Score the model")

        _hlayout_1 = Container(
            layout="horizontal",
            widgets=[
                self._fit_button,
                self._predict_button,
                self._fit_predict_button,
            ],
        )
        _hlayout_1.margins = (0, 0, 0, 0)
        _hlayout_2 = Container(
            layout="horizontal",
            widgets=[
                self._transform_button,
                self._describe_button,
                self._score_button,
            ],
        )
        _hlayout_2.margins = (0, 0, 0, 0)

        self._output_area = TextEdit(label="Output")
        self._output_area.max_height = 100
        self._output_area.read_only = True
        self._output_area.native.setFontFamily(MonospaceFontFamily)

        super().__init__(
            widgets=[
                self._model_widget,
                self._data_widget,
                _hlayout_1,
                _hlayout_2,
                self._output_area,
            ],
            **kwargs,
        )
        self.margins = (0, 0, 0, 0)

        # connect signals
        self._model_widget.changed.connect(self._on_model_changed)
        self._fit_button.changed.connect(self._fit)
        self._predict_button.changed.connect(self._predict)
        self._fit_predict_button.changed.connect(self._fit_predict)
        self._transform_button.changed.connect(self._transform)
        self._describe_button.changed.connect(self._describe)
        self._score_button.changed.connect(self._score)

    @classmethod
    def new(cls) -> SkLearnContainer:
        """Build a new widget (singleton constructor)."""
        if cls._current_widget is None:
            cls._current_widget = SkLearnContainer()
        return cls._current_widget

    def _on_model_changed(self, model: SkLearnModelProtocol):
        """Check attributes of the model and enable/disable buttons."""
        self._predict_button.enabled = hasattr(model, "predict")
        self._fit_predict_button.enabled = hasattr(model, "fit_predict")
        self._transform_button.enabled = hasattr(model, "transform")
        self._score_button.enabled = hasattr(model, "score")

    def _fit(self):
        """Run self.fit(X, Y)."""
        table = find_current_table(self)
        input = self._data_widget.get_values(table.data)
        self._model_widget.model.fit(input.X, input.Y)
        text: str = self._model_widget._text.value
        if not text.endswith(" (fitted)"):
            self._model_widget._text.value = text + " (fitted)"

    def _predict(self):
        """Run self.predict(X)."""
        table = find_current_table(self)
        input = self._data_widget.get_values(table.data)
        predicted = self._model_widget.model.predict(input.X)
        name = table.columns.coerce_name("predicted")
        table.assign({name: predicted})
        table.selections = [(slice(None), table.columns.get_loc(name))]

        # update Y if it is empty
        if input.Y is None:
            self._data_widget._Y_widget._read_selection(table)
        return None

    def _fit_predict(self):
        """Run self.fit_predict(X, Y)."""
        table = find_current_table(self)
        input = self._data_widget.get_values(table.data)
        predicted = self._model_widget.model.fit_predict(input.X, input.Y)
        text: str = self._model_widget._text.value
        if not text.endswith(" (fitted)"):
            self._model_widget._text.value = text + " (fitted)"
        name = table.columns.coerce_name("predicted")
        table.assign({name: predicted})
        table.selections = [(slice(None), table.columns.get_loc(name))]

        # update Y if it is empty
        if input.Y is None:
            self._data_widget._Y_widget._read_selection(table)
        return None

    def _transform(self):
        """Run self.transform(X)."""
        table = find_current_table(self)
        input = self._data_widget.get_values(table.data)
        transformed = self._model_widget.model.transform(input.X)
        df_dict = {}
        for i in range(transformed.shape[1]):
            name = table.columns.coerce_name("transformed", start=i)
            df_dict[name] = transformed[:, i]
        nc_before = len(table.columns)
        table.assign(df_dict)
        nc_after = len(table.columns)
        table.selections = [(slice(None), slice(nc_before, nc_after))]
        return None

    def _describe(self):
        """Describe current model state."""
        model = self._model_widget.model
        output: list[str] = []
        params = model.get_params()
        output.append(f"parameters = \n{params}")

        for k, v in vars(model).items():
            if k.endswith("_"):
                output.append(f"{k} = \n{v}")

        self._output_area.value = "\n\n".join(output)
        return None

    def _score(self):
        """Run self.score(X)."""
        table = find_current_table(self)
        input = self._data_widget.get_values(table.data)
        score = self._model_widget.model.score(input.X, input.Y)
        self._output_area.value = f"score = {score}"



from __future__ import annotations
from functools import reduce
from operator import ior
from typing import Any, TYPE_CHECKING
from qtpy import QtWidgets as QtW
from qtpy.QtCore import Qt

from ._titlebar import QTitleBar

if TYPE_CHECKING:
    from ._mainwindow import QMainWindow


class QtDockWidget(QtW.QDockWidget):
    areas = {
        "left": Qt.DockWidgetArea.LeftDockWidgetArea,
        "right": Qt.DockWidgetArea.RightDockWidgetArea,
        "top": Qt.DockWidgetArea.TopDockWidgetArea,
        "bottom": Qt.DockWidgetArea.BottomDockWidgetArea,
    }

    def __init__(
        self,
        parent: QMainWindow,
        widget: QtW.QWidget,
        *,
        name: str = "",
        area: str = "right",
        allowed_areas: list[str] = None,
    ):
        super().__init__(name, parent)

        # set title bar
        _titlebar = QTitleBar(name, self)
        self.setTitleBarWidget(_titlebar)
        _titlebar.setCursor(Qt.CursorShape.OpenHandCursor)
        _titlebar.closeSignal.connect(self.close)

        areas = self.__class__.areas

        if allowed_areas:
            if not isinstance(allowed_areas, (list, tuple)):
                raise TypeError("`allowed_areas` must be a list or tuple")

            if any(area not in areas for area in allowed_areas):
                raise ValueError(
                    f"all allowed_areas argument must be in {set(areas.keys())}"
                )
            allowed_areas = reduce(ior, [areas[a] for a in allowed_areas])
        else:
            allowed_areas = Qt.DockWidgetArea.AllDockWidgetAreas
        self.qt_area = areas[area]
        self.setAllowedAreas(allowed_areas)
        self.setWidget(widget)
        self.setMinimumHeight(50)
        self.setMinimumWidth(50)

    def setSourceObject(self, obj):
        self._widget = obj

    @property
    def widget(self) -> QtW.QWidget | Any:
        return self._widget

    # fmt: off
    if TYPE_CHECKING:
        def parentWidget(self) -> QMainWindow: ...
    # fmt: on



from __future__ import annotations
from typing import TYPE_CHECKING, Callable, Iterable, Iterator, NamedTuple
from contextlib import contextmanager
from psygnal import Signal

from tabulous._range import TableAnchorBase, translate_slice

if TYPE_CHECKING:
    Range = tuple[slice, slice]


class Index(NamedTuple):
    """Index tuple,"""

    row: int
    column: int

    def as_uint(self) -> Index:
        """Return unsigned Index"""
        return Index(max(self.row, 0), max(self.column, 0))


class DummyRange(NamedTuple):
    row: slice
    column: slice


_DUMMY_RANGE = DummyRange(slice(0, 0), slice(0, 0))


class RangesModel(TableAnchorBase):
    """Custom 2D range model for efficient overlay handling on a large table."""

    def __init__(self):
        self._ranges: list[Range] = []
        self._is_blocked = False
        self._selected_indices: set[int] = set()
        self._row_selection_indices: set[int] = set()
        self._col_selection_indices: set[int] = set()

    def __len__(self) -> int:
        """Number of ranges"""
        return len(self._ranges)

    def __iter__(self) -> Iterator[Range]:
        """Iterate over all the ranges."""
        return iter(self._ranges)

    def __getitem__(self, index: int) -> Range:
        """Get the range at the specified index."""
        return self._ranges[index]

    @property
    def ranges(self) -> list[Range]:
        return list(self._ranges)

    def iter_row_selections(self) -> Iterator[slice]:
        for i in self._row_selection_indices:
            yield self._ranges[i][0]

    def iter_col_selections(self) -> Iterator[slice]:
        for i in self._col_selection_indices:
            yield self._ranges[i][1]

    def as_ranges(self) -> list[Range]:
        """Return a list of ranges considering row/column selections."""
        out: list[Range] = []
        for i, rng in enumerate(self._ranges):
            if i in self._row_selection_indices:
                if i in self._col_selection_indices:
                    out.append((slice(None), slice(None)))
                else:
                    out.append((rng[0], slice(None)))
            elif i in self._col_selection_indices:
                out.append((slice(None), rng[1]))
            else:
                out.append(rng)
        return out

    def append(self, range: Range, row: bool = False, column: bool = False) -> None:
        """Append a new range."""
        if self._is_blocked:
            return None
        self._ranges.append(range)
        if row:
            self._row_selection_indices.add(len(self._ranges) - 1)
        elif column:
            self._col_selection_indices.add(len(self._ranges) - 1)
        return None

    def extend(
        self, ranges: Iterable[Range], row: bool = False, column: bool = False
    ) -> None:
        """Extend the ranges with a list of ranges."""
        if self._is_blocked:
            return None
        len0 = len(self._ranges)
        self._ranges.extend(ranges)
        len1 = len(self._ranges)
        if row:
            self._row_selection_indices.update(range(len0, len1))
        elif column:
            self._col_selection_indices.update(range(len0, len1))
        return None

    def update_last(self, range: Range, row: bool = False, col: bool = False) -> None:
        """Update the last range with new one."""
        if self._is_blocked:
            return None
        if self._ranges:
            self._ranges[-1] = range
        else:
            self._ranges.append(range)
        if row:
            self._row_selection_indices.add(len(self._ranges) - 1)
        elif col:
            self._col_selection_indices.add(len(self._ranges) - 1)
        return None

    def set_ranges(self, ranges: list[Range]) -> None:
        if self._is_blocked:
            return None
        self.clear()
        return self._ranges.extend(ranges)

    def clear(self) -> None:
        """Clear all the selections"""
        if self._is_blocked:
            return None
        self._ranges.clear()
        self._row_selection_indices.clear()
        self._col_selection_indices.clear()
        return None

    @contextmanager
    def blocked(self) -> None:
        """Block selection updates in this context."""
        self._is_blocked = True
        try:
            yield
        finally:
            self._is_blocked = False

    def reorder_to_last(self, idx: int):
        rng = self._ranges.pop(idx)
        self._ranges.append(rng)
        if idx in self._row_selection_indices:
            self._row_selection_indices.discard(idx)
            self._row_selection_indices.add(len(self._ranges) - 1)
        elif idx in self._col_selection_indices:
            self._col_selection_indices.discard(idx)
            self._col_selection_indices.add(len(self._ranges) - 1)
        return None

    def select(self, indices: Iterable[int]) -> None:
        self._selected_indices = set(indices)
        return None

    def add_selection(self, index: int) -> None:
        nranges = len(self._ranges)
        if index < 0:
            index += nranges
        if 0 <= index < nranges:
            self._selected_indices.add(index)
        else:
            raise ValueError(f"Index {index} out of range")
        return None

    def delete_selected(self):
        remain = [
            self._ranges[i]
            for i in range(len(self._ranges))
            if i not in self._selected_indices
        ]
        self._ranges = remain
        self._row_selection_indices -= self._selected_indices
        self._col_selection_indices -= self._selected_indices
        self._selected_indices = set()
        return None

    def iter_ranges_under_index(
        self,
        row: int,
        col: int,
        *,
        reverse: bool = True,
    ) -> Iterator[tuple[int, Range]]:
        """Iterate over all the ranges that are under the specified position."""
        # If current index is on the headers, simply check using the adjacent cell
        if row == -1:
            row = 0
        if col == -1:
            col = 0

        if reverse:
            rmax = len(self._ranges) - 1
            for i, (rr, cc) in enumerate(reversed(self._ranges)):
                if rr.start <= row < rr.stop and cc.start <= col < cc.stop:
                    yield rmax - i, (rr, cc)
        else:
            for i, (rr, cc) in enumerate(self._ranges):
                if rr.start <= row < rr.stop and cc.start <= col < cc.stop:
                    yield i, (rr, cc)

    def range_under_index(self, row: int, col: int) -> tuple[int, Range | None]:
        """Get the last range that is under the specified position."""
        try:
            out = next(self.iter_ranges_under_index(row, col))
        except StopIteration:
            out = -1, None
        return out

    def insert_rows(self, row: int, count: int) -> None:
        for i, (r, c) in enumerate(self._ranges):
            r = translate_slice(r, row, count)
            self._ranges[i] = (r, c)

    def insert_columns(self, col: int, count: int) -> None:
        for i, (r, c) in enumerate(self._ranges):
            c = translate_slice(c, col, count)
            self._ranges[i] = (r, c)

    def remove_rows(self, row: int, count: int) -> None:
        to_be_removed = []
        for i, (r, c) in enumerate(self._ranges):
            r = translate_slice(r, row, -count)
            if r.start >= r.stop:
                to_be_removed.append(i)
            self._ranges[i] = (r, c)
        for i in reversed(to_be_removed):
            self._ranges.pop(i)

    def remove_columns(self, col: int, count: int) -> None:
        to_be_removed = []
        for i, (r, c) in enumerate(self._ranges):
            c = translate_slice(c, col, -count)
            if c.start >= c.stop:
                to_be_removed.append(i)
            self._ranges[i] = (r, c)
        for i in reversed(to_be_removed):
            self._ranges.pop(i)


class SelectionModel(RangesModel):
    """A specialized range model with item-selection-like behavior."""

    moving = Signal(Index, Index)
    moved = Signal(Index, Index)

    def __init__(self, row_count: Callable[[], int], col_count: Callable[[], int]):
        super().__init__()
        self._ctrl_on = False
        self._shift_on = False
        self._selection_start: Index | None = None
        self._current_index = Index(0, 0)
        self._row_count_getter = row_count
        self._col_count_getter = col_count

    @property
    def current_index(self) -> Index:
        """Current position of the selection cursor."""
        return self._current_index

    @current_index.setter
    def current_index(self, index: tuple[int, int]):
        self._current_index = Index(*index)

    @property
    def current_range(self) -> tuple[slice, slice] | None:
        if len(self._ranges) > 0:
            return self._ranges[-1]
        return None

    def iter_all_indices(self) -> Iterator[tuple[int, int]]:
        """Iterate all the indices (int, int) in all the selection ranges."""
        nr = self._row_count_getter()
        nc = self._col_count_getter()
        for rng in self.ranges:
            for r in range(*rng[0].indices(nr)):
                for c in range(*rng[1].indices(nc)):
                    yield r, c

    @property
    def start(self) -> Index | None:
        """The selection starting index."""
        return self._selection_start

    def is_jumping(self) -> bool:
        """Whether the selection is jumping or not."""
        return len(self._ranges) > 0 and self._ranges[-1] is _DUMMY_RANGE

    def is_moving_to_edge(self) -> bool:
        """Whether the selection is moving to the edge by Ctrl+arrow key."""
        return not self.is_jumping() and self._ctrl_on

    def set_ctrl(self, on: bool) -> None:
        """Equivalent to pressing Ctrl."""
        self._ctrl_on = bool(on)
        return None

    def set_shift(self, on: bool) -> None:
        """Equivalent to pressing Shift."""
        self._shift_on = bool(on)
        if on and self._selection_start is None:
            self._selection_start = self._current_index
        return None

    def reset(self) -> None:
        self._selection_start = None

    def jump_to(self, r: int, c: int):
        """Emulate mouse click at cell (r, c)."""
        if self._ctrl_on and not self._shift_on:
            self._ranges.append(_DUMMY_RANGE)
        return self.move_to(r, c)

    def move_to(self, r: int, c: int):
        """Emulate dragging to cell (r, c)."""
        src = self._current_index
        dst = Index(r, c)
        self.moving.emit(src, dst)
        self._current_index = dst
        if self._is_blocked:
            return None

        if not self._shift_on:
            self._selection_start = None

        if self._selection_start is None:
            _r0 = _r1 = r
            _c0 = _c1 = c
        else:
            r0, c0 = self._selection_start
            _r0, _r1 = sorted([r0, r])
            _c0, _c1 = sorted([c0, c])

        if _r0 < 0:
            rsl = slice(0, self._row_count_getter())
            col = True
        else:
            rsl = slice(_r0, _r1 + 1)
            col = False
        if _c0 < 0:
            csl = slice(0, self._col_count_getter())
            row = True
        else:
            csl = slice(_c0, _c1 + 1)
            row = False

        if not self._shift_on:
            if not self.is_jumping():
                self.clear()
        elif self._selection_start is None:
            self._selection_start = self._current_index

        self.update_last((rsl, csl), row=row, col=col)
        self.moved.emit(src, dst)
        return None

    def move(self, dr: int, dc: int, allow_header: bool = False):
        """Move by (dr, dc) cells."""
        r, c = self._current_index
        idx_min = -int(allow_header)

        if dr != 0:
            nr = self._row_count_getter()
            r = min(r + dr, nr - 1)
            r = max(idx_min, r)

        if dc != 0:
            nc = self._col_count_getter()
            c = min(c + dc, nc - 1)
            c = max(idx_min, c)

        return self.move_to(r, c)

    def insert_rows(self, row: int, count: int) -> None:
        super().insert_rows(row, count)
        _r, _c = self._current_index
        if row <= _r:
            self.current_index = (_r + count, _c)

    def insert_columns(self, col: int, count: int) -> None:
        super().insert_columns(col, count)
        _r, _c = self._current_index
        if col <= _c:
            self.current_index = (_r, _c + count)

    def remove_rows(self, row: int, count: int) -> None:
        super().remove_rows(row, count)
        _r, _c = self._current_index
        if row + count <= _r:
            self.current_index = (_r - count, _c)
        elif row <= _r:
            self.current_index = (row, _c)

    def remove_columns(self, col: int, count: int) -> None:
        super().remove_columns(col, count)
        _r, _c = self._current_index
        if col + count <= _c:
            self.current_index = (_r, _c - count)
        elif col <= _c:
            self.current_index = (_r, col)



from __future__ import annotations

from typing import TYPE_CHECKING
from io import StringIO

if TYPE_CHECKING:
    from tabulous.widgets import TableViewerBase
    from tabulous.widgets._table import _DataFrameTableLayer


# This identifier is inaccessible from the console because it contains dots.
# It is only used from the local namespace dict.
VIEWER_IDENTIFIER = "__ipython.tabulous.viewer__"


def _get_viewer(ns: dict) -> TableViewerBase:
    """Return the viewer from the namespace."""
    viewer = ns.get(VIEWER_IDENTIFIER, None)
    if viewer is None:
        raise RuntimeError("Viewer not found in namespace")
    return viewer


def _filter_args(line: str) -> tuple[list[str], list[str]]:
    quot_flag_1 = False
    quot_flag_2 = False
    args: list[str] = []
    keywords: list[str] = []
    idx = 0
    for i, c in enumerate(line + " "):
        if c == " ":
            if not (quot_flag_1 or quot_flag_2) and (word := line[idx:i]):
                if word.startswith("-"):
                    keywords.append(word)
                else:
                    args.append(word)
                idx = i + 1
        elif c == '"':
            quot_flag_2 = not quot_flag_2
        elif c == "'":
            quot_flag_1 = not quot_flag_1

    return args, keywords


_INSTALLED = False


def install_magics():
    # magic commands can only be registered within a score where `get_ipython` is
    # available.
    global _INSTALLED

    if _INSTALLED:
        return

    from IPython import get_ipython  # noqa: F401
    from IPython.core.magic import (
        register_line_magic,
        register_cell_magic,
        register_line_cell_magic,
        needs_local_scope,
    )

    @register_line_magic
    @needs_local_scope
    def add(line: str, local_ns: dict = {}):
        """
        Add an object to the table viewer.

        Examples
        --------
        >>> %add df
        >>> %add df as table
        """
        viewer = _get_viewer(local_ns)
        if " as " in line:
            obj, typ = line.rsplit(" as ", maxsplit=1)
        else:
            obj = line.strip()
            typ = "table"
        if typ == "table":
            _add_fn = viewer.add_table
        elif typ == "spreadsheet":
            _add_fn = viewer.add_spreadsheet
        elif typ == "groupby":
            _add_fn = viewer.add_groupby
        elif typ == "loader":
            _add_fn = viewer.add_loader
        else:
            raise ValueError(
                f"Unknown type {typ}. Must be one of 'table', "
                "'spreadsheet', 'groupby', or 'loader'."
            )
        ns = local_ns.copy()
        ns["__builtins__"] = {}
        return _add_fn(eval(obj, ns, {}))

    @register_cell_magic
    @needs_local_scope
    def csv(line: str, cell: str | None = None, local_ns: dict = {}):
        """
        Convert a CSV string to a spreadsheet.

        Examples
        --------
        >>> viewer.add_spreadsheet(
        ...     {"a": [1, 4], "b": [2, 5], "c": [3, 6]},
        ...     name="my_table"
        ... )

        is identical to

        >>> %%table my_table
        >>> a b c
        >>> 1 2 3
        >>> 4 5 6

        """
        import pandas as pd

        if cell is None:
            raise ValueError("No input object provided")
        buf = StringIO(cell)
        df = pd.read_csv(buf)
        viewer = _get_viewer(local_ns)
        return viewer.add_spreadsheet(df, name=line)

    @register_line_magic
    @needs_local_scope
    def filter(line: str, local_ns: dict = {}):
        """
        Query-style filtration on a table.

        Examples
        --------
        >>> %filter a > 3  # equivalent to df.proxy.filter("a > 3")
        """
        viewer = _get_viewer(local_ns)
        if line == "":
            return viewer.current_table.proxy.filter(None)
        return viewer.current_table.proxy.filter(line, local_ns)

    @register_line_cell_magic
    @needs_local_scope
    def query(line: str, cell: str | None = None, local_ns: dict = {}):
        """
        Query-style evaluation on a table

        Examples
        --------
        >>> %query b = a + 1

        >>> %%query
        >>> b = a + 1
        >>> c = b > 3
        """
        viewer = _get_viewer(local_ns)
        table = viewer.current_table
        if table.table_type not in ("Table", "SpreadSheet"):
            raise ValueError("Querying is only supported for Tables and SpreadSheets")
        table: _DataFrameTableLayer
        if cell is not None:
            lines = cell.splitlines()
        else:
            lines = [line]
        for line in lines:
            line = line.strip()
            if line == "":
                continue
            table.query(line)

    @register_line_magic
    @needs_local_scope
    def plot(line: str, local_ns: dict = {}):
        viewer = _get_viewer(local_ns)
        table = viewer.current_table
        args, keys = _filter_args(line)
        if "-n" in keys or "--new" in keys:
            table.plt.new_widget()
        if "-c" in keys or "--clear" in keys:
            table.plt.cla()
        if len(args) == 0:
            return
        elif len(args) == 1:
            y = table[args[0]].data
            table.plt.plot(y)
        else:
            x = table[args[0]].data
            for yname in args[1:]:
                y = table[yname].data
                table.plt.plot(x, y)
        return

    @register_line_magic
    @needs_local_scope
    def scatter(line: str, local_ns: dict = {}):
        viewer = _get_viewer(local_ns)
        table = viewer.current_table
        args, keys = _filter_args(line)
        if "-n" in keys or "--new" in keys:
            table.plt.new_widget()
        if "-c" in keys or "--clear" in keys:
            table.plt.cla()
        if len(args) == 0:
            return
        elif len(args) == 1:
            raise ValueError("scatter requires at least two arguments (X and Y)")
        else:
            x = table[args[0]].data
            for yname in args[1:]:
                y = table[yname].data
                table.plt.scatter(x, y)
        return

    del add, csv, filter, query, plot, scatter
    _INSTALLED = True
    return None



from __future__ import annotations
from typing import Callable, TypeVar

_T = TypeVar("_T", bound=Callable)

# fmt: off
_PARAMETERS = dict(
    name="name : str, optional\n\tName of the table.",
    editable="editable : bool, default is False\n\tWhether the table is editable via UI.",
    copy="copy : bool, default is True\n\tWhether to copy the data before adding to avoid overwriting the original one.",
    metadata="metadata : dict, optional\n\tMetadata of the table.",
    update="update : bool, default is False\n\tIf True, update the table data if a table of same name exists.",
)
# fmt: on


def update_doc(f: _T) -> _T:
    """Update the docstring of the given object."""
    doc = f.__doc__
    if doc:
        doc_lines = doc.splitlines()
        indent = doc_lines[1][: -len(doc_lines[1].lstrip())]
        doc = doc.replace("}{", f"}}\n{indent}{{")
        params = {k: _expand_indent(v, indent) for k, v in _PARAMETERS.items()}
        doc = doc.format(**params)
    f.__doc__ = doc
    return f


def _expand_indent(s: str, indent: str) -> str:
    line0, line1 = s.split("\n")
    return f"{line0}\n{indent}{line1}"



from __future__ import annotations
from typing import TYPE_CHECKING, Callable
from qtpy import QtWidgets as QtW
from qtpy.QtCore import Qt, QTimer, Signal
from ._base import QBaseTable, _QTableViewEnhanced, DataFrameModel

if TYPE_CHECKING:
    import pandas as pd


def _get_standard_icon(x):
    return QtW.QApplication.style().standardIcon(x)


class _QPlayButton(QtW.QToolButton):
    """The play/pause button."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self._RUN = _get_standard_icon(QtW.QStyle.StandardPixmap.SP_MediaPlay)
        self._PAUSE = _get_standard_icon(QtW.QStyle.StandardPixmap.SP_MediaPause)
        self.clicked.connect(self.switchRunning)
        self.setRunning(True)

    def running(self) -> bool:
        return self._running

    def setRunning(self, val: bool):
        self._running = val
        if val:
            self.setIcon(self._PAUSE)
        else:
            self.setIcon(self._RUN)
        return None

    def switchRunning(self, *_):
        return self.setRunning(not self.running())

    def copy(self, link: bool = True) -> _QPlayButton:
        new = self.__class__(self.parent())
        new.setRunning(self.running())
        if link:
            new.clicked.connect(self.click)
        return new


class _QTimerSpinBox(QtW.QSpinBox):
    timeout = Signal()

    def __init__(self, parent=None, value: int = 100):
        super().__init__(parent)
        self.setMinimum(10)
        self.setMaximum(10000)
        self.setSuffix(" ms")

        self._qtimer = QTimer()
        self._qtimer.setSingleShot(True)
        self._qtimer.setInterval(value)
        self._qtimer.setTimerType(Qt.TimerType.PreciseTimer)
        self.valueChanged.connect(self._qtimer.setInterval)
        self.valueChanged.connect(self._set_step)
        self._qtimer.timeout.connect(self.timeout.emit)
        self.setValue(value)

    def copy(self, link: bool = True) -> _QTimerSpinBox:
        new = self.__class__(value=self.value())
        if link:
            self.valueChanged.connect(new.setValue)
            new.valueChanged.connect(self.setValue)
        return new

    def start(self):
        return self._qtimer.start()

    def stop(self):
        return self._qtimer.stop()

    def toggleTimer(self):
        if self._qtimer.isActive():
            self._qtimer.stop()
        else:
            self._qtimer.start()
        return None

    def _set_step(self, val: int):
        nd = len(str(val))
        if nd == 2:
            step = 10
        else:
            step = 5 * 10 ** (nd - 2)
        self.setSingleStep(step)
        return None


class _QTableDisplayWidget(QtW.QWidget):
    _table_display: QTableDisplay | None = None
    _timer: _QTimerSpinBox | None = None
    _play_btn: _QPlayButton | None = None
    _qtable_view: _QTableViewEnhanced | None = None

    @classmethod
    def from_table_display(cls, table_display: QTableDisplay) -> _QTableDisplayWidget:
        """Create a new QTableDisplayWidget from an existing QTableDisplay."""
        self = cls.from_widgets(
            spinbox=table_display._timer,
            play_button=table_display._play_btn,
            qtable_view=table_display._qtable_view_,
        )

        table_display.loaded.connect(self._on_loaded)
        self._table_display = table_display

        return self

    def copy(self, link: bool = True) -> _QTableDisplayWidget:
        new = self.from_widgets(
            spinbox=self._table_display._timer.copy(link=link),
            play_button=self._table_display._play_btn.copy(link=link),
            qtable_view=self._table_display._qtable_view_.copy(link=link),
        )
        new._table_display = self._table_display
        if link:
            self._table_display.loaded.connect(new._on_loaded)
        return new

    @classmethod
    def from_widgets(
        cls,
        spinbox: _QTimerSpinBox,
        play_button: _QPlayButton,
        qtable_view: _QTableViewEnhanced,
    ) -> _QTableDisplayWidget:
        """Create a new QTableDisplay from existing widgets."""
        self = cls()
        self._timer = spinbox
        self._play_btn = play_button
        self._qtable_view = qtable_view

        _header = QtW.QWidget()
        _header_layout = QtW.QHBoxLayout()
        _header_layout.setContentsMargins(2, 2, 2, 2)
        _header_layout.addWidget(QtW.QLabel("Interval:"))

        _header_layout.addWidget(spinbox)
        _header_layout.addWidget(play_button)
        _header.setLayout(_header_layout)

        _main_layout = QtW.QVBoxLayout()
        _main_layout.setContentsMargins(0, 0, 0, 0)
        _main_layout.addWidget(_header)
        _main_layout.addWidget(qtable_view)
        self.setLayout(_main_layout)

        return self

    def _on_loaded(self):
        if self._play_btn.running():
            self._timer.start()
        else:
            self._timer.stop()
        return self._qtable_view._update_all()

    @property
    def _selection_model(self):
        return self._qtable_view._selection_model

    @property
    def _on_moving(self):
        return self._qtable_view._on_moving

    @property
    def _on_moved(self):
        return self._qtable_view._on_moved


# TODO: don't initialize filter and only accept function filter.
class QTableDisplay(QBaseTable):
    loaded = Signal()

    def __init__(
        self,
        parent: QtW.QWidget | None = None,
        loader: Callable[[], pd.DataFrame] = None,
        interval_ms: int = 1000,
    ):
        import pandas as pd

        self._timer = _QTimerSpinBox(value=interval_ms)
        self._play_btn = _QPlayButton()
        self._play_btn.clicked.connect(self._timer.toggleTimer)

        super().__init__(parent, pd.DataFrame([]))
        if loader is None:
            self._loader = lambda: pd.DataFrame([])
        else:
            self._loader = lambda: pd.DataFrame(loader())
        self._loading = False
        self._timer.timeout.connect(self._on_timeout)

        if self._play_btn.running():
            self._timer.start()

    if TYPE_CHECKING:

        def model(self) -> DataFrameModel:
            ...

    def _on_timeout(self):
        """Run refresh if needed."""
        if self._play_btn.running() and not self._loading:
            self._load_data()

    def loader(self) -> Callable:
        """Return the loader function."""
        return self._loader

    def setLoader(self, loader: Callable) -> None:
        """Set the loader function and refresh."""
        if not callable(loader):
            raise TypeError("loader must be callable")
        self._loader = loader
        return self._load_data()

    def getDataFrame(self) -> pd.DataFrame:
        return self._data_raw

    def setDataFrame(self, data: pd.DataFrame) -> None:
        self._data_raw = data
        self.model().df = data
        self._qtable_view.viewport().update()
        self._filtered_index = data.index
        self._filtered_columns = data.columns
        return

    def createModel(self):
        model = DataFrameModel(self)
        self._qtable_view.setModel(model)
        return None

    @property
    def _qtable_view(self) -> _QTableViewEnhanced:
        return self._qtable_view_

    @property
    def _central_widget(self) -> _QTableDisplayWidget:
        return self._central_widget_

    def createQTableView(self):
        self._qtable_view_ = _QTableViewEnhanced(self)

        wdt = _QTableDisplayWidget.from_table_display(self)
        self.addWidget(wdt)
        self._central_widget_ = wdt

    def _load_data(self) -> None:
        self._loading = True
        if self.isVisible():
            try:
                self._data_raw = self._loader()
            except Exception as e:
                return
            self.model().df = self._data_raw
            self._filtered_index = self._data_raw.index
            self._filtered_columns = self._data_raw.columns

        self._loading = False
        self.loaded.emit()
        return None

    def running(self) -> bool:
        """True if the loader is running."""
        return self._play_btn.running()

    def setRunning(self, running: bool) -> None:
        """Set the loader running state."""
        self._play_btn.setRunning(running)
        return None

    def interval(self) -> int:
        return self._timer._qtimer.interval()

    def setInterval(self, interval: int) -> None:
        return self._timer._qtimer.setInterval(interval)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Hashable
import re
from io import StringIO
import warnings

import numpy as np
import pandas as pd
from qtpy import QtCore, QtGui
from qtpy.QtCore import Qt

from collections_undo import arguments

from ._base import AbstractDataFrameModel, QMutableSimpleTable
from ._animation import RowAnimation, ColumnAnimation
from tabulous._dtype import get_converter, get_dtype, DTypeMap, DefaultValidator
from tabulous._utils import TabulousConfig, get_config
from tabulous.color import normalize_color
from tabulous.types import ItemInfo
from tabulous._text_formatter import DefaultFormatter
from tabulous import _pd_index

if TYPE_CHECKING:
    from magicgui.widgets._bases import ValueWidget

# More rows/columns will be displayed
_STRING_DTYPE = get_dtype("string")
_EMPTY = object()
_EXP_FLOAT = re.compile(r"[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)")


class SpreadSheetModel(AbstractDataFrameModel):
    """A DataFrameModel for a spreadsheet."""

    def __init__(self, parent=None):
        super().__init__(parent)
        from tabulous._utils import get_config

        self._table_config = get_config().table
        self._columns_dtype = self.parent()._columns_dtype
        self._out_of_bound_color_cache = None

    @property
    def _out_of_bound_color(self) -> QtGui.QColor:
        if self._out_of_bound_color_cache is not None:
            return self._out_of_bound_color_cache
        qtable_view = self.parent()._qtable_view
        if viewer := qtable_view.parentViewer():
            bgcolor = viewer.backgroundColor()
        else:
            bgcolor = qtable_view.palette().color(qtable_view.backgroundRole())
        r, g, b = bgcolor.red(), bgcolor.green(), bgcolor.blue()
        if r + b + g > 382.5:
            qcolor = QtGui.QColor(max(r - 4, 0), max(g - 4, 0), b)
        else:
            qcolor = QtGui.QColor(r, g, max(b + 4, 0))
        self._out_of_bound_color_cache = qcolor
        return qcolor

    @property
    def df(self) -> pd.DataFrame:  # NOTE: this returns a string data frame
        return self._df

    @df.setter
    def df(self, data: pd.DataFrame):
        self._df = data

    def rowCount(self, parent=None):
        return self._table_config.max_row_count

    def columnCount(self, parent=None):
        return self._table_config.max_column_count

    def _data_display(self, index: QtCore.QModelIndex):
        """Display role."""
        r, c = index.row(), index.column()
        df = self.df
        if r < df.shape[0] and c < df.shape[1]:
            val = df.iat[r, c]
            colname = df.columns[c]
            if mapper := self._text_formatter.get(colname, None):
                _converter = get_converter(
                    self._columns_dtype.get(colname, _STRING_DTYPE)
                )
                try:
                    text = str(mapper(_converter(val)))
                except Exception:
                    text = self._FORMAT_ERROR
            else:
                text = str(val)
            # Exponentially formatted float numbers are not displayed correctly.
            if _EXP_FLOAT.match(text):
                text = format(float(text), ".5e")
            return text
        return QtCore.QVariant()

    def _data_background_color(self, index: QtCore.QModelIndex):
        r, c = index.row(), index.column()
        df = self.df
        if r < df.shape[0] and c < df.shape[1]:
            colname = df.columns[c]
            if mapper := self._background_colormap.get(colname, None):
                val = df.iat[r, c]
                try:
                    col = mapper(val)
                    if col is None:
                        return QtCore.QVariant()
                    rgba = normalize_color(col)
                except Exception as e:
                    # since this method is called many times, errorous function should
                    # be deleted from the mapper.
                    self._background_colormap.pop(colname)
                    raise e
                return QtGui.QColor(*rgba)
        else:
            return self._out_of_bound_color  # add shade to the out-of-range cells

    def _column_tooltip(self, section: int):
        name = self.df.columns[section]
        if dtype := self._columns_dtype.get(name, None):
            return f"{name} (dtype: {dtype})"
        else:
            dtype = self.df.dtypes.values[section]
            return f"{name} (dtype: infer)"

    def _data_tooltip(self, index: QtCore.QModelIndex):
        r, c = index.row(), index.column()
        if r < self.df.shape[0] and c < self.df.shape[1]:
            val = self.df.iat[r, c]
            name = self.df.columns[c]
            qtable = self.parent()
            r = qtable._proxy.get_source_index(r)
            if slot := qtable._qtable_view._table_map.get_by_dest((r, c), None):
                ref = f"\nExpr: {slot.as_literal(dest=True)}"
                if slot._current_error is not None:
                    ref += "\n" + slot.format_error()
            else:
                ref = ""
            dtype = self._columns_dtype.get(name, None)
            if dtype is None:
                return f"{val!r} (dtype: infer){ref}"
            else:
                return f"{val!r} (dtype: {dtype}){ref}"
        return QtCore.QVariant()

    def rename_column(self, old_name: str, new_name: str):
        super().rename_column(old_name, new_name)
        if dtype := self._columns_dtype.pop(old_name, None):
            self._columns_dtype[new_name] = dtype
        return None

    if TYPE_CHECKING:

        def parent(self) -> QSpreadSheet:
            ...


class QSpreadSheet(QMutableSimpleTable):
    """
    A table layer class that works similar to Excel sheet.

    Unlike ``QTableLayer``, this class does not have dtype. The dtype will be
    determined every time table data is converted into DataFrame. Table data
    is (almost) unbounded.
    """

    _DEFAULT_EDITABLE = True
    NaN = ""

    def __init__(self, parent=None, data: pd.DataFrame | None = None):
        from tabulous._utils import get_config

        cfg = get_config()

        if data is not None:
            if (
                cfg.table.max_row_count < data.shape[0]
                or cfg.table.max_column_count < data.shape[1]
            ):
                _max_size = (cfg.table.max_row_count, cfg.table.max_column_count)
                raise ValueError(
                    f"Input table data size {data.shape} exceeds the maximum "
                    f"size {_max_size}."
                )

        self._columns_dtype = DTypeMap()
        super().__init__(parent, data)
        self._qtable_view.verticalHeader().setMinimumWidth(20)
        animate = cfg.window.animate
        self._anim_row = RowAnimation(self).set_animate(animate)
        self._anim_col = ColumnAnimation(self).set_animate(animate)

        # initialize section spans
        rspan, cspan = cfg.table.row_size, cfg.table.column_size
        nr, nc = self._data_raw.shape
        self._qtable_view.verticalHeader().insertSection(0, nr, rspan)
        self._qtable_view.horizontalHeader().insertSection(0, nc, cspan)

    if TYPE_CHECKING:

        def model(self) -> SpreadSheetModel:
            ...

    def getDataFrame(self) -> pd.DataFrame:
        if self._data_cache is not None:
            return self._data_cache
        # Convert table data into a DataFrame with the optimal dtypes
        _sep = "\t"
        data_raw = self._data_raw
        if data_raw.shape[1] > 0:
            val = data_raw.to_csv(sep=_sep, index=False)
            buf = StringIO(val)
            out: pd.DataFrame = pd.read_csv(
                buf,
                sep=_sep,
                header=0,
                na_values=["#ERROR"],
                names=data_raw.columns,
                **self._columns_dtype.as_pandas_kwargs(),
            )
            out.index = data_raw.index
            out.columns = data_raw.columns
        else:
            out = pd.DataFrame(index=data_raw.index, columns=[])
        self._data_cache = out
        return out

    def _get_sub_frame(self, columns: list[str]) -> pd.DataFrame:
        """Parse and return a sub-frame of the table."""
        if self._data_cache is not None:
            return self._data_cache[columns]
        is_series = not isinstance(columns, list)
        if is_series:
            columns = [columns]
        _sep = "\t"
        data_raw = self._data_raw[columns]
        _dtype_map = self._columns_dtype.copy()
        _to_be_deleted = []
        for key in _dtype_map:
            if key not in columns:
                _to_be_deleted.append(key)
        for key in _to_be_deleted:
            del _dtype_map[key]

        val = data_raw.to_csv(sep=_sep, index=False)
        buf = StringIO(val)
        out: pd.DataFrame = pd.read_csv(
            buf,
            sep=_sep,
            header=0,
            na_values=["#ERROR"],
            names=data_raw.columns,
            **_dtype_map.as_pandas_kwargs(),
        )
        out.index = data_raw.index
        if is_series:
            out = out.iloc[:, 0]
        return out

    def dataShape(self) -> tuple[int, int]:
        """Shape of data."""
        return self.model().df.shape

    def dataShown(self, parse: bool = False) -> pd.DataFrame:
        """Return the shown dataframe (consider filter)."""
        if parse:
            df = self.getDataFrame()
            return self._proxy.apply(df)
        else:
            return self.model().df

    def load_config(self, cfg: TabulousConfig | None = None):
        """Load new config."""
        self._anim_row.set_animate(cfg.window.animate)
        self._anim_col.set_animate(cfg.window.animate)
        return super().load_config(cfg)

    @QMutableSimpleTable._mgr.interface
    def setDataFrame(self, data: pd.DataFrame) -> None:
        """Set data frame as a string table."""
        self._data_raw = data.astype(_STRING_DTYPE)

        # SpreadSheet columns should be str if possible. Convert it.
        if isinstance(self._data_raw.columns, pd.RangeIndex):
            self._data_raw.columns = _pd_index.char_arange(self._data_raw.columns.size)
        elif self._data_raw.columns.dtype.kind in "iuf":
            self._data_raw.columns = self._data_raw.columns.astype(str)

        self._data_cache = None
        self.setProxy(None)
        self.refreshTable()
        return

    def updateValue(self, r, c, val):
        index = self._data_raw.index[r]
        columns = self._data_raw.columns[c]
        # NOTE: It seems very weird but the string array of pandas does not
        # support setting (N, 1) string array.
        if isinstance(val, pd.DataFrame):
            # NOTE loc-indexer takes axes into consideration. Here, input data
            # frame needs to be updated.
            val.index = index
            val.columns = columns
            if isinstance(c, slice) and c.stop == 1:
                val = pd.Series(val.iloc[:, 0], dtype="string")

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._data_raw.loc[index, columns] = val
        if self._proxy.proxy_type != "none":
            self._set_proxy(self._proxy)
        return self.refreshTable()

    @setDataFrame.server
    def setDataFrame(self, data):
        return arguments(getattr(self, "_data_raw", None))

    @setDataFrame.set_formatter
    def _setDataFrame_fmt(self, data: pd.DataFrame):
        return f"set new data of shape {data.shape}"

    def _apply_proxy(self):
        if self._proxy.proxy_type == "none":
            return self.tableSlice()
        return self._proxy.apply(self.tableSlice(), ref=self.getDataFrame)

    __delete = object()

    @QMutableSimpleTable._mgr.interface
    def assignColumns(self, serieses: dict[str, pd.Series]):
        to_delete = set()
        to_assign: dict[str, pd.Series] = {}
        for k, v in serieses.items():
            if v is self.__delete:
                to_delete.add(k)
            else:
                to_assign[k] = v.astype(_STRING_DTYPE)
        old_value = self._data_raw
        self._data_raw: pd.DataFrame = self._data_raw.assign(**to_assign).drop(
            to_delete, axis=1, inplace=False
        )

        self.model().df = self._data_raw
        self.setProxy(None)
        self.refreshTable()
        # NOTE: ItemInfo cannot have list indices.
        self.itemChangedSignal.emit(
            ItemInfo(
                slice(None),
                slice(None),
                self._data_raw,
                old_value,
            )
        )
        self._data_cache = None
        self._edited = True
        return None

    @assignColumns.server
    def assignColumns(self, serieses: dict[str, pd.Series]):
        columns = self._data_raw.columns
        old_param: dict[str, pd.Series] = {}
        for k in serieses.keys():
            if k in columns:
                old_param[k] = self._data_raw[k]
            else:
                old_param[k] = self.__delete
        return arguments(old_param)

    @assignColumns.set_formatter
    def _assignColumns_fmt(self, serieses: dict[str, pd.Series]):
        keys = list(serieses.keys())
        return f"assign new data to {keys}"

    def createModel(self) -> None:
        """Create spreadsheet model."""
        model = SpreadSheetModel(self)
        self._qtable_view.setModel(model)
        return None

    def convertValue(self, c: int, value: Any) -> Any:
        """Convert value to the type of the table."""
        return value

    def readClipBoard(self, sep=r"\s+"):
        """Read clipboard as a string data frame."""
        return pd.read_clipboard(
            header=None, sep=sep, dtype=_STRING_DTYPE
        )  # read as string

    def setDataFrameValue(self, r: int | slice, c: int | slice, value: Any) -> None:
        nr, nc = self._data_raw.shape
        rmax = _get_limit(r)
        cmax = _get_limit(c)
        need_expand = nr <= rmax or nc <= cmax

        if isinstance(value, str):
            if need_expand and value == "":
                # if user start editing a cell outside the data frame and did nothing,
                # do not expand the data frame.
                return
            if isinstance(r, int) and isinstance(c, int) and value == "NA":
                # if user start editing an empty cell and did nothing, don't set "NA".
                model = self._qtable_view.model()
                index = model.index(r, c, QtCore.QModelIndex())
                text = model.data(index, Qt.ItemDataRole.DisplayRole)
                if text == value:
                    return

        elif isinstance(value, pd.DataFrame) and any(value.dtypes != "string"):
            value = value.astype(_STRING_DTYPE)

        with self._mgr.merging(formatter=lambda _: self._set_value_fmt(r, c, value)):
            if need_expand:
                self.expandDataFrame(max(rmax - nr + 1, 0), max(cmax - nc + 1, 0))
            super().setDataFrameValue(r, c, value)
            self._set_proxy(self._proxy)

        return None

    def setLabeledData(self, r: slice, c: slice, value: pd.Series):
        nr, nc = self._data_raw.shape
        rmax = _get_limit(r)
        cmax = _get_limit(c)
        need_expand = nr <= rmax or nc <= cmax

        if value.dtype != "string":
            value = value.astype(_STRING_DTYPE)

        with self._mgr.merging(formatter=lambda cmds: cmds[-2].format()):
            if need_expand:
                self.expandDataFrame(max(rmax - nr + 1, 0), max(cmax - nc + 1, 0))
            super().setLabeledData(r, c, value)
            self._set_proxy(self._proxy)

        return None

    def _pre_set_array(self, r: slice, c: slice, _value: pd.DataFrame):
        """Convert input dataframe for setting to data[r, c]."""
        if len(self.model()._validator) == 0:
            # use faster method if no validator is set
            out = self._data_raw.iloc[r, c].copy()
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                if _value.size == 1:
                    val = _value.values[0, 0]
                    if val == "NA":
                        val = ""
                    out.loc[:, :] = val
                else:
                    out.loc[:, :] = _value.astype(_STRING_DTYPE).fillna("").values
            return out
        return super()._pre_set_array(r, c, _value)

    @QMutableSimpleTable._mgr.undoable
    def expandDataFrame(self, nrows: int, ncols: int):
        """Expand the data frame by adding empty rows and columns."""
        if not self.isEditable():
            return None
        nr, nc = self._data_raw.shape
        self._data_raw = _pad_dataframe(self._data_raw, nrows, ncols)
        if self._proxy.proxy_type != "none":
            self._set_proxy(self._proxy)  # need update!
        cfg = get_config()

        rspan, cspan = cfg.table.row_size, cfg.table.column_size
        self._qtable_view.verticalHeader().insertSection(nr, nrows, rspan)
        self._qtable_view.horizontalHeader().insertSection(nc, ncols, cspan)
        return None

    @expandDataFrame.undo_def
    def expandDataFrame(self, nrows: int, ncols: int):
        nr, nc = self._data_raw.shape
        self._data_raw = self._data_raw.iloc[: nr - nrows, : nc - ncols]
        self._set_proxy(self._proxy)

        self._qtable_view.verticalHeader().removeSection(nr, nrows)
        self._qtable_view.horizontalHeader().insertSection(nc, ncols)
        self._data_cache = None
        return None

    @QMutableSimpleTable._mgr.undoable
    def insertRows(
        self, row: int, count: int, value: Any = _EMPTY, span: int | None = None
    ):
        """Insert rows at the given row number and count."""
        if self._proxy.proxy_type != "none":
            raise NotImplementedError("Cannot insert rows during filtering/sorting.")

        index_existing = self._data_raw.index

        if value is _EMPTY:
            # determine index labels
            index: list[int] = []
            i = 0
            while True:
                if i not in index_existing:
                    index.append(i)
                    if len(index) >= count:
                        break
                else:
                    i += 1

            value = _df_full(
                nrows=count,
                ncols=self._data_raw.shape[1],
                index=index,
                columns=self._data_raw.columns,
            )

        self._data_raw = pd.concat(
            [
                self._data_raw.iloc[:row, :],
                value,
                self._data_raw.iloc[row:, :],
            ],
            axis=0,
        )
        if _pd_index.is_ranged(index_existing):
            self._data_raw.index = pd.RangeIndex(0, self._data_raw.index.size)
        self.model().insertRows(row, count, QtCore.QModelIndex(), span=span)

        @self._anim_row.connect
        def _on_finish():
            self._set_proxy(self._proxy)
            self._data_cache = None
            self._edited = True

            # update indices
            self._process_insert_rows(row, count)

            info = ItemInfo(
                slice(row, row + count),
                slice(None),
                value,
                ItemInfo.INSERTED,
            )
            self._qtable_view.setZoom(self._qtable_view.zoom())
            self.itemChangedSignal.emit(info)

        return self._anim_row.run_insert(row, count, span)

    @insertRows.undo_def
    def insertRows(self, row: int, count: int, value: Any = _EMPTY, span=None):
        """Insert rows at the given row number and count."""
        return self.removeRows(row, count)

    @insertRows.set_formatter
    def _insertRows_fmt(self, row: int, count: int, value: Any = _EMPTY):
        return f"table.index.insert(at={row}, count={count})"

    def insertColumns(
        self, col: int, count: int, value: Any = _EMPTY, span: int | None = None
    ):
        """Insert columns at the given column number and count."""
        with self._mgr.merging(
            lambda cmds: f"table.columns.insert(at={col}, count={count})"
        ):
            self._insert_columns(col, count, value, span)
            self._process_header_widgets_on_insert(col, count)
        return None

    @QMutableSimpleTable._mgr.undoable
    def _insert_columns(
        self, col: int, count: int, value: Any = _EMPTY, span: int | None = None
    ):
        columns_existing = self._data_raw.columns
        _c_ranged = _pd_index.is_ranged(columns_existing)

        if value is _EMPTY:
            columns = _remove_duplicate(
                _pd_index.char_arange(col, col + count), existing=columns_existing
            )
            value = _df_full(
                nrows=self._data_raw.shape[0],
                ncols=count,
                index=self._data_raw.index,
                columns=columns,
            )

        self._data_raw = pd.concat(
            [
                self._data_raw.iloc[:, :col],
                value,
                self._data_raw.iloc[:, col:],
            ],
            axis=1,
        )
        # if column index is ranged chars, re-assign them
        if _c_ranged:
            self._data_raw.columns = _pd_index.char_range_index(
                self._data_raw.columns.size
            )
            model = self.model()
            for d in [
                model._foreground_colormap,
                model._background_colormap,
                model._text_formatter,
                model._validator,
                self._columns_dtype,
            ]:
                _inserte_in_dict(d, col, count)

        self.model().insertColumns(col, count, QtCore.QModelIndex(), span=span)

        @self._anim_col.connect
        def _on_finish():
            self._data_cache = None
            self._edited = True

            # update indices
            self._process_insert_columns(col, count)

            info = ItemInfo(
                slice(None),
                slice(col, col + count),
                value,
                ItemInfo.INSERTED,
            )
            self._qtable_view.setZoom(self._qtable_view.zoom())
            self.itemChangedSignal.emit(info)

        return self._anim_col.run_insert(col, count, span)

    @_insert_columns.undo_def
    def _insert_columns(self, col: int, count: int, value: Any = _EMPTY, span=None):
        """Insert columns at the given column number and count."""
        self._remove_columns(col, count, self._data_raw.iloc[:, col : col + count])
        self._set_proxy(self._proxy)
        return None

    def removeRows(self, row: int, count: int):
        """Remove rows at the given row number and count."""
        df = self.model().df.iloc[row : row + count, :]
        hheader = self._qtable_view.verticalHeader()
        spans = hheader._section_sizes[row : row + count].copy()
        with self._mgr.merging(
            lambda cmds: f"table.index.remove(at={row}, count={count})"
        ):
            self._clear_incell_slots(
                slice(row, row + count),
                slice(0, self._data_raw.shape[1]),
            )
            self._remove_rows(row, count, df, spans)
        return None

    @QMutableSimpleTable._mgr.undoable
    def _remove_rows(self, row: int, count: int, old_values: pd.DataFrame, spans=None):
        _r_ranged = isinstance(self._data_raw.index, pd.RangeIndex)
        _new_data_raw = pd.concat(
            [self._data_raw.iloc[:row, :], self._data_raw.iloc[row + count :, :]],
            axis=0,
        )
        if _r_ranged:
            _new_data_raw.index = pd.RangeIndex(0, _new_data_raw.index.size)

        @self._anim_row.connect
        def _on_finish():
            self._data_raw = _new_data_raw
            self.model().removeRows(row, count, QtCore.QModelIndex())
            self._set_proxy(self._proxy)
            self.setSelections(
                [(slice(row, row + 1), slice(0, self._data_raw.shape[1]))]
            )
            self._data_cache = None
            self._edited = True

            self._process_remove_rows(row, count)
            info = ItemInfo(
                slice(row, row + count), slice(None), ItemInfo.DELETED, old_values
            )
            self._qtable_view.setZoom(self._qtable_view.zoom())
            self.itemChangedSignal.emit(info)

        self._anim_row.run_remove(row, count)
        return None

    @_remove_rows.undo_def
    def _remove_rows(self, row: int, count: int, old_values: pd.DataFrame, spans=None):
        self.insertRows(row, count, old_values, spans)
        self.setSelections([(slice(row, row + 1), slice(0, self._data_raw.shape[1]))])
        return None

    def removeColumns(self, column: int, count: int):
        """Remove columns at the given column number and count."""
        df = self.model().df.iloc[:, column : column + count]
        with self._mgr.merging(
            lambda cmds: f"table.columns.remove(at={column}, count={count})"
        ):
            self._clear_incell_slots(
                slice(0, self._data_raw.shape[0]),
                slice(column, column + count),
            )
            self._process_header_widgets_on_remove(column, count)
            model = self.model()
            hheader = self._qtable_view.horizontalHeader()
            for index in range(column, column + count):
                colname = model.df.columns[index]
                self.setForegroundColormap(colname, None)
                self.setBackgroundColormap(colname, None)
                self.setTextFormatter(colname, None)
                self.setDataValidator(colname, None)
                self.setColumnDtype(colname, None)
            section_sizes = hheader._section_sizes[column : column + count].copy()
            self._remove_columns(column, count, df, section_sizes)
        return None

    @QMutableSimpleTable._mgr.undoable
    def _remove_columns(
        self,
        col: int,
        count: int,
        old_values: pd.DataFrame,
        spans: np.ndarray | None = None,
    ):
        _c_ranged = _pd_index.is_ranged(self._data_raw.columns)
        _new_data_raw = pd.concat(
            [self._data_raw.iloc[:, :col], self._data_raw.iloc[:, col + count :]],
            axis=1,
        )
        if _c_ranged:
            _new_data_raw.columns = _pd_index.char_range_index(
                _new_data_raw.columns.size
            )
            model = self.model()
            for d in [
                model._foreground_colormap,
                model._background_colormap,
                model._text_formatter,
                model._validator,
                self._columns_dtype,
            ]:
                _remove_in_dict(d, col, count)

        @self._anim_col.connect
        def _on_finish():
            self.model().removeColumns(col, count, QtCore.QModelIndex())
            self._data_raw = _new_data_raw
            self.setSelections(
                [(slice(0, self._data_raw.shape[0]), slice(col, col + 1))]
            )
            self._data_cache = None
            self._edited = True

            self._process_remove_columns(col, count)
            self._qtable_view.setZoom(self._qtable_view.zoom())
            info = ItemInfo(
                slice(None), slice(col, col + count), ItemInfo.DELETED, old_values
            )
            self.itemChangedSignal.emit(info)  # relay to the psygnal.Signal
            self._set_proxy(self._proxy)
            return

        self._anim_col.run_remove(col, count)
        return None

    @_remove_columns.undo_def
    def _remove_columns(
        self,
        col: int,
        count: int,
        old_values: pd.DataFrame,
        spans: np.ndarray | None = None,
    ):
        self.insertColumns(col, count, old_values, spans)
        self.setSelections([(slice(0, self._data_raw.shape[0]), slice(col, col + 1))])
        return None

    @QMutableSimpleTable._mgr.interface
    def _set_widget_at_index(self, r: int, c: int, widget: ValueWidget | None) -> None:
        index = self.model().index(r, c)
        if wdt := self._qtable_view.indexWidget(index):
            try:
                self.itemChangedSignal.disconnect(wdt._tabulous_callback)
            except TypeError:
                pass
            wdt.close()

        if widget is None:
            # equivalent to delete the widget
            return None

        if widget.widget_type in ("CheckBox", "RadioButton"):

            def converter(x):
                return x != "False"

        elif widget.widget_type in ("SpinBox", "Slider"):
            converter = int
        elif widget.widget_type in ("FloatSpinBox", "FloatSlider"):
            converter = float
        else:
            converter = str

        def _sig():
            with widget.changed.blocked():
                val = self.model().df.iat[r, c]
                try:
                    widget.value = converter(val)
                except Exception:
                    self.setDataFrameValue(r, c, str(widget.value))
                    raise

        if self.model().df.iat[r, c] != "":
            _sig()
        widget.native._tabulous_callback = _sig
        self.itemChangedSignal.connect(_sig)
        widget.changed.connect(lambda val: self.setDataFrameValue(r, c, str(val)))
        self._qtable_view.setIndexWidget(index, widget.native)

    @_set_widget_at_index.server
    def _set_widget_at_index(self, r: int, c: int, widget: ValueWidget):
        index = self.model().index(r, c)
        wdt = self._qtable_view.indexWidget(index)
        return arguments(r, c, getattr(wdt, "_magic_widget", None))

    def setVerticalHeaderValue(self, index: int, value: Any) -> None:
        """Set value of the table vertical header and DataFrame at the index."""
        if value == "":
            return
        nrows = self._data_raw.shape[0]

        with self._mgr.merging(formatter=lambda cmds: cmds[-1].format()):
            if index >= nrows:
                self.expandDataFrame(index - nrows + 1, 0)
            self._set_proxy(self._proxy)
            super().setVerticalHeaderValue(index, value)
            self._data_cache = None

        return None

    def setHorizontalHeaderValue(self, index: int, value: Any) -> None:
        """Set value of the table horizontal header and DataFrame at the index."""
        if value == "":
            return
        ncols = self._data_raw.shape[1]

        with self._mgr.merging(formatter=lambda cmds: cmds[-1].format()):
            if index >= ncols:
                self.expandDataFrame(0, index - ncols + 1)
            old_name = self._data_raw.columns[index]
            self._set_proxy(self._proxy)
            super().setHorizontalHeaderValue(index, value)
            if old_name in self._columns_dtype.keys():
                self.setColumnDtype(value, self._columns_dtype.pop(old_name))
            self._data_cache = None

        return None

    @QMutableSimpleTable._mgr.interface
    def setColumnDtype(self, label: Hashable, dtype: Any | None) -> None:
        """Set the dtype of the column with the given label."""
        if dtype is None:
            # delete cache if dtype used to be set
            if self._columns_dtype.pop(label, None):
                self._data_cache = None
        else:
            if label not in self._data_raw.columns:
                raise ValueError(f"Column {label!r} not found.")

            dtype = get_dtype(dtype)
            if self._columns_dtype.get(label, None) is not dtype:
                self._columns_dtype[label] = dtype
                self._data_cache = None

        if validator := self.model()._validator.get(label, None):
            if isinstance(validator, DefaultValidator):
                self.model()._validator.pop(label)
        if formatter := self.model()._text_formatter.get(label, None):
            if isinstance(formatter, DefaultFormatter):
                self.model()._text_formatter.pop(label)
        return None

    @setColumnDtype.server
    def setColumnDtype(self, label: Hashable, dtype: Any):
        return arguments(label, self._columns_dtype.get(label, None))

    def _set_default_data_validator(self, name: Hashable):
        """Set default data validator based on the dtype."""
        dtype = self._columns_dtype[name]
        validator = DefaultValidator(dtype)
        return self.setDataValidator(name, validator)

    def _set_default_text_formatter(self, name: Hashable):
        """Set default data formatter based on the dtype."""
        dtype = self._columns_dtype[name]
        formatter = DefaultFormatter(dtype)
        return self.setTextFormatter(name, formatter)

    def _process_header_widgets_on_insert(self, column: int, count: int):
        to_increment: list[int] = []
        header_widgets = self._header_widgets().copy()
        for k in header_widgets.keys():
            if k >= column:
                to_increment.append(k)
        to_increment.sort(reverse=True)
        for k in to_increment:
            header_widgets[k + count] = header_widgets.pop(k)
        self.updateHorizontalHeaderWidget(header_widgets)
        self._set_proxy(self._proxy)
        return None

    def _process_header_widgets_on_remove(self, column: int, count: int):
        to_remove: list[int] = []
        to_decrement: list[int] = []
        header_widgets = self._header_widgets().copy()
        for k in header_widgets.keys():
            if k >= column:
                if k < column + count:
                    to_remove.append(k)
                else:
                    to_decrement.append(k)
        for k in to_remove:
            header_widgets.pop(k)
        to_decrement.sort()
        for k in to_decrement:
            header_widgets[k - count] = header_widgets.pop(k)
        self.updateHorizontalHeaderWidget(header_widgets)
        return None


def _pad_dataframe(df: pd.DataFrame, nr: int, nc: int, value: Any = "") -> pd.DataFrame:
    """Pad a dataframe by nr rows and nr columns with the given value."""
    if df.shape == (0, 0):
        return _df_full(nr, nc, value, columns=_pd_index.char_range_index(nc))

    # pad rows
    _nr, _nc = df.shape
    _r_ranged = isinstance(df.index, pd.RangeIndex)
    _c_ranged = _pd_index.is_ranged(df.columns)
    if nr > 0:
        # find unique index
        if df.index.size == 0:
            index = range(nr)
        else:
            index = range(_nr, _nr + nr)
        ext = _df_full(nr, _nc, value, index=index, columns=df.columns)
        df = pd.concat([df, ext], axis=0)
        if _r_ranged:
            df.index = pd.RangeIndex(0, df.index.size)

    # pad columns
    _nr, _nc = df.shape  # NOTE: shape may have changed
    if nc > 0:
        columns = _pd_index.char_arange(df.columns.size, df.columns.size + nc)
        # check duplication
        if not _c_ranged:
            columns = _remove_duplicate(columns, existing=df.columns)
        ext = _df_full(_nr, nc, value, index=df.index, columns=columns)
        df = pd.concat([df, ext], axis=1)
        if _c_ranged:
            df.columns = _pd_index.char_range_index(df.columns.size)
    return df


def _get_limit(a) -> int:
    if isinstance(a, int):
        amax = a
    elif isinstance(a, slice):
        amax = a.stop - 1
    else:
        raise TypeError(f"Cannot infer limit of type {type(a)}")
    return amax


def _df_full(
    nrows: int, ncols: int, value="", index=None, columns=None
) -> pd.DataFrame:
    """A DataFrame filled with the given value."""
    return pd.DataFrame(
        np.full((nrows, ncols), value),
        index=index,
        columns=columns,
        dtype="string",
    )


def _remove_duplicate(
    columns: np.ndarray,
    existing: pd.Index,
    copy: bool = False,
) -> np.ndarray:
    """Remove names from `columns` that have duplicates in `existing`"""
    if copy:
        columns = columns.copy()
    for ic, c in enumerate(columns):
        if c in existing:
            i = 0
            c0 = f"{c}_{i}"
            while c0 in existing:
                i += 1
                c0 = f"{c}_{i}"
            columns[ic] = c0
    return columns


def _inserte_in_dict(d: dict[str, Any], start: int, count: int) -> None:
    out: dict[str, Any] = {}
    for k, v in d.items():
        if _pd_index.str_to_num(k) >= start:
            out[_pd_index.increment(k, count)] = v
        else:
            out[k] = v
    d.clear()
    d.update(out)
    return None


def _remove_in_dict(d: dict[str, Any], start: int, count: int) -> None:
    out: dict[str, Any] = {}
    for k, v in d.items():
        if _pd_index.str_to_num(k) >= start + count:
            out[_pd_index.decrement(k, count)] = v
        else:
            out[k] = v
    d.clear()
    d.update(out)
    return None



from __future__ import annotations
from typing import Iterable, TYPE_CHECKING, Union
import numpy as np
from tabulous.color import ColorTuple, normalize_color
from tabulous.types import ColorType
from tabulous._dtype import isna, get_converter

if TYPE_CHECKING:
    from pandas.core.dtypes.dtypes import CategoricalDtype
    import pandas as pd

    _TimeLike = Union[pd.Timestamp, pd.Timedelta]
    _ColorType = tuple[int, int, int, int]

_DEFAULT_MIN = "#697FD1"
_DEFAULT_MAX = "#FF696B"


def exec_colormap_dialog(ds: pd.Series, parent=None):
    """Open a dialog to define a colormap for a series."""
    from tabulous._magicgui import ColorEdit
    from magicgui.widgets import Dialog, LineEdit, Container

    dtype = ds.dtype
    if dtype == "category":
        dtype: CategoricalDtype
        widgets = [
            ColorEdit(value=_random_color(), label=str(cat)) for cat in dtype.categories
        ]
        dlg = Dialog(widgets=widgets)
        dlg.native.setParent(parent, dlg.native.windowFlags())
        if dlg.exec():
            return dict(zip(dtype.categories, (w.value for w in widgets)))

    elif dtype.kind in "uif":  # unsigned int, int, float
        lmin = LineEdit(value=str(ds.min()))
        lmax = LineEdit(value=str(ds.max()))
        cmin = ColorEdit(value=_DEFAULT_MIN)
        cmax = ColorEdit(value=_DEFAULT_MAX)
        min_ = Container(
            widgets=[cmin, lmin], labels=False, layout="horizontal", label="Min"
        )
        min_.margins = (0, 0, 0, 0)
        max_ = Container(
            widgets=[cmax, lmax], labels=False, layout="horizontal", label="Max"
        )
        max_.margins = (0, 0, 0, 0)
        dlg = Dialog(widgets=[min_, max_])
        dlg.native.setParent(parent, dlg.native.windowFlags())
        if dlg.exec():
            return segment_by_float(
                [(float(lmin.value), cmin.value), (float(lmax.value), cmax.value)]
            )

    elif dtype.kind == "b":  # boolean
        false_ = ColorEdit(value=_DEFAULT_MIN, label="False")
        true_ = ColorEdit(value=_DEFAULT_MAX, label="True")
        dlg = Dialog(widgets=[false_, true_])
        dlg.native.setParent(parent, dlg.native.windowFlags())
        if dlg.exec():
            converter = get_converter(np.dtype(bool))
            _dict = {False: false_.value, True: true_.value}
            return lambda val: _dict.get(converter(val), None)

    elif dtype.kind in "mM":  # time stamp or time delta
        min_ = ColorEdit(value=_DEFAULT_MIN, label="Min")
        max_ = ColorEdit(value=_DEFAULT_MAX, label="Max")
        dlg = Dialog(widgets=[min_, max_])
        dlg.native.setParent(parent, dlg.native.windowFlags())
        if dlg.exec():
            return segment_by_time(
                [(ds.min(), min_.value), (ds.max(), max_.value)], dtype
            )

    else:
        raise NotImplementedError(
            f"Dtype {dtype!r} not supported. Please set colormap programmatically."
        )


def _random_color() -> list[int]:
    return list(np.random.randint(256, size=3)) + [255]


def _where(x, border: Iterable[float]) -> int:
    for i, v in enumerate(border):
        if x < v:
            return i - 1
    return len(border) - 1


def segment_by_float(maps: list[tuple[float, ColorType]]):
    converter = get_converter(np.dtype("f"))
    borders: list[float] = []
    colors: list[ColorTuple] = []
    for v, c in maps:
        borders.append(v)
        colors.append(normalize_color(c))
    idx_max = len(borders) - 1

    # check is sorted
    if not all(borders[i] <= borders[i + 1] for i in range(len(borders) - 1)):
        raise ValueError("Borders must be sorted")

    def _colormap(value: float) -> _ColorType:
        if isna(value):
            return None
        value = converter(value)
        idx = _where(value, borders)
        if idx == -1 or idx == idx_max:
            return colors[idx]
        min_color = np.array(colors[idx], dtype=np.float64)
        max_color = np.array(colors[idx + 1], dtype=np.float64)
        min = borders[idx]
        max = borders[idx + 1]
        return (value - min) / (max - min) * (max_color - min_color) + min_color

    return _colormap


def segment_by_time(maps: list[tuple[_TimeLike, ColorType]], dtype):
    converter = get_converter(dtype)
    borders: list[_TimeLike] = []
    colors: list[ColorTuple] = []
    for v, c in maps:
        borders.append(v)
        colors.append(normalize_color(c))
    idx_max = len(borders) - 1

    # check is sorted
    if not all(borders[i] <= borders[i + 1] for i in range(len(borders) - 1)):
        raise ValueError("Borders must be sorted")

    def _colormap(value: _TimeLike) -> _ColorType:
        if isna(value):
            return None
        value = converter(value)
        idx = _where(value, borders)
        if idx == -1 or idx == idx_max:
            return colors[idx]
        min_color = np.array(colors[idx], dtype=np.float64)
        max_color = np.array(colors[idx + 1], dtype=np.float64)
        min = borders[idx].value
        max = borders[idx + 1].value
        return (value.value - min) / (max - min) * (max_color - min_color) + min_color

    return _colormap



from __future__ import annotations

from qtpy import QtWidgets as QtW, QtGui, QtCore
from qtpy.QtCore import Qt, Property, Signal
from enum import Enum


class ButtonState(Enum):
    NONE = "none"
    SQUARE = "square"
    TRIANGLE = "triangle"


class QCircularProgressBar(QtW.QWidget):
    abortRequested = Signal()
    infiniteRequested = Signal(bool)

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._value = 0
        self._minimum = 0
        self._maximum = 100
        self._timer = QtCore.QTimer(self)
        self._radius = 20
        self._barWidth = 4
        self._infinite = False
        self._pen = QtGui.QPen(
            QtGui.QColor(0, 48, 48, 255),
            self._barWidth,
            Qt.PenStyle.SolidLine,
            Qt.PenCapStyle.RoundCap,
        )
        self._groove_pen = QtGui.QPen(
            QtGui.QColor(230, 230, 230, 255),
            self._barWidth,
            Qt.PenStyle.SolidLine,
            Qt.PenCapStyle.RoundCap,
        )
        self._btn_state = ButtonState.NONE
        self.setMouseTracking(True)
        self.infiniteRequested.connect(self.setInfinite)

    def sizeHint(self) -> QtCore.QSize:
        return QtCore.QSize(self._radius * 2, self._radius * 2)

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        painter = QtGui.QPainter(self)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing, True)
        painter.setRenderHint(QtGui.QPainter.RenderHint.TextAntialiasing, True)
        center = self.rect().center()
        x0 = center.x() - self._radius
        y0 = center.y() - self._radius
        rect = QtCore.QRect(x0, y0, 2 * self._radius, 2 * self._radius)
        ratio = self._value / self._maximum
        min_ratio = self._minimum / self._maximum
        _full = 360 * 16
        painter.setPen(self._groove_pen)
        painter.drawArc(rect, 0, _full)
        painter.setPen(self._pen)
        painter.drawArc(rect, 90 * 16 - int(_full * min_ratio), -int(_full * ratio))

        if self._btn_state is ButtonState.SQUARE:
            _l = self._radius * 0.4
            x0 = center.x() - _l
            y0 = center.y() - _l
            rect = QtCore.QRectF(x0, y0, 2 * _l, 2 * _l)
            painter.setPen(QtGui.QPen(self._pen.color(), 1))
            painter.setBrush(QtGui.QBrush(self._pen.color()))
            painter.drawRect(rect)
        elif self._btn_state is ButtonState.TRIANGLE:
            path = QtGui.QPainterPath()
            _l = self._radius * 0.6
            path.moveTo(center.x() - _l / 2, center.y() + _l * 0.866)
            path.lineTo(center.x() - _l / 2, center.y() - _l * 0.866)
            path.lineTo(center.x() + _l, center.y())
            path.lineTo(center.x() - _l / 2, center.y() + _l * 0.866)
            painter.fillPath(path, QtGui.QBrush(self._pen.color()))

    def value(self) -> float:
        return self._value

    def setValue(self, value: float) -> None:
        if value < 0:
            self.infiniteRequested.emit(True)
        else:
            self.infiniteRequested.emit(False)
            self._value = value
            self.update()

    @Property(QtGui.QColor)
    def grooveColor(self) -> QtGui.QColor:
        return self._groove_pen.color()

    @grooveColor.setter
    def grooveColor(self, color: QtGui.QColor) -> None:
        self._groove_pen.setColor(color)
        self.update()

    @Property(QtGui.QColor)
    def color(self) -> QtGui.QColor:
        return self._pen.color()

    @color.setter
    def color(self, color: QtGui.QColor) -> None:
        self._pen.setColor(color)
        self.update()

    def buttonState(self) -> ButtonState:
        return self._btn_state

    def setButtonState(self, state: ButtonState) -> None:
        self._btn_state = ButtonState(state)
        self.update()

    def _is_inside(self, pos: QtCore.QPoint) -> bool:
        center = self.rect().center()
        dist = (center.x() - pos.x()) ** 2 + (center.y() - pos.y()) ** 2
        return dist < self._radius**2

    def mouseMoveEvent(self, event: QtGui.QMouseEvent) -> None:
        if self._is_inside(event.pos()):
            self.setCursor(Qt.CursorShape.PointingHandCursor)
        else:
            self.setCursor(Qt.CursorShape.ArrowCursor)
        return super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:
        if event.button() == Qt.MouseButton.LeftButton and self._is_inside(event.pos()):
            self.abortRequested.emit()

    def setInfinite(self, infinite: bool) -> None:
        if infinite and not self._infinite:
            self._timer.setInterval(15)
            self._timer.timeout.connect(self._on_infinite_timeout)
            self._timer.start()
            self._minimum = 0
            self._value = 25
        elif not infinite and self._infinite:
            self._timer.stop()
            self._value = 0
            self._minimum = 0
        self._infinite = infinite
        self.update()

    def _on_infinite_timeout(self):
        self._minimum = (self._minimum + 1) % self._maximum
        self.update()

    def radius(self) -> float:
        """Radius of the progress bar in pixels"""
        return self._radius

    def setRadius(self, radius: float) -> None:
        """Set radius of the progress bar in pixels"""
        self._radius = radius
        self.update()

    def barWidth(self) -> float:
        return self._barWidth

    def setBarWidth(self, width: float) -> None:
        """Set width of the progress bar in pixels"""
        self._barWidth = width
        self._pen.setWidth(width)
        self._groove_pen.setWidth(width)
        self.update()



__version__ = "0.5.3"

from tabulous.widgets import TableViewer, TableViewerWidget
from tabulous.core import (
    current_viewer,
    read_csv,
    read_excel,
    view_table,
    view_spreadsheet,
    open_sample,
)
from tabulous._magicgui import MagicTableViewer

__all__ = [
    "TableViewer",
    "TableViewerWidget",
    "MagicTableViewer",
    "current_viewer",
    "read_csv",
    "read_excel",
    "view_table",
    "view_spreadsheet",
    "open_sample",
]
